# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numbers

import numpy as np
from scipy.sparse import csc_matrix, csr_matrix, issparse
from sklearn.base import BaseEstimator
from sklearn.ensemble._base import BaseEnsemble
from sklearn.ensemble._gb import BaseGradientBoosting, VerboseReporter
from sklearn.ensemble._gradient_boosting import _random_sample_mask
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeRegressor
from sklearn.tree._tree import DTYPE
from sklearn.utils import check_random_state
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.extmath import squared_norm
from sklearn.utils.validation import (
    _check_sample_weight,
    check_array,
    check_is_fitted,
)

from ..base import SurvivalAnalysisMixin
from ..linear_model.coxph import BreslowEstimator
from ..util import check_array_survival
from .survival_loss import (
    LOSS_FUNCTIONS,
    CensoredSquaredLoss,
    CoxPH,
    IPCWLeastSquaresError,
)

__all__ = [
    "ComponentwiseGradientBoostingSurvivalAnalysis",
    "GradientBoostingSurvivalAnalysis",
]


def _sample_binomial_plus_one(p, size, random_state):
    drop_model = random_state.binomial(1, p=p, size=size)
    n_dropped = np.sum(drop_model)
    if n_dropped == 0:
        idx = random_state.randint(0, size)
        drop_model[idx] = 1
        n_dropped = 1
    return drop_model, n_dropped


class _ComponentwiseLeastSquares(BaseEstimator):
    def __init__(self, component):
        self.component = component

    def fit(self, X, y, sample_weight):
        xw = X[:, self.component] * sample_weight
        b = np.dot(xw, y)
        if b == 0:
            self.coef_ = 0
        else:
            a = np.dot(xw, xw)
            self.coef_ = b / a

        return self

    def predict(self, X):
        return X[:, self.component] * self.coef_


def _fit_stage_componentwise(X, residuals, sample_weight, **fit_params):  # pylint: disable=unused-argument
    """Fit component-wise weighted least squares model"""
    n_features = X.shape[1]

    base_learners = []
    error = np.empty(n_features)
    for component in range(n_features):
        learner = _ComponentwiseLeastSquares(
            component).fit(X, residuals, sample_weight)
        l_pred = learner.predict(X)
        error[component] = squared_norm(residuals - l_pred)
        base_learners.append(learner)

    # TODO: could use bottleneck.nanargmin for speed
    best_component = np.nanargmin(error)
    best_learner = base_learners[best_component]
    return best_learner


class ComponentwiseGradientBoostingSurvivalAnalysis(BaseEnsemble, SurvivalAnalysisMixin):
    r"""Gradient boosting with component-wise least squares as base learner.

    See the :ref:`User Guide </user_guide/boosting.ipynb>` and [1]_ for further description.

    Parameters
    ----------
    loss : {'coxph', 'squared', 'ipcwls'}, optional, default: 'coxph'
        loss function to be optimized. 'coxph' refers to partial likelihood loss
        of Cox's proportional hazards model. The loss 'squared' minimizes a
        squared regression loss that ignores predictions beyond the time of censoring,
        and 'ipcwls' refers to inverse-probability of censoring weighted least squares error.

    learning_rate : float, optional, default: 0.1
        learning rate shrinks the contribution of each base learner by `learning_rate`.
        There is a trade-off between `learning_rate` and `n_estimators`.
        Values must be in the range `[0.0, inf)`.

    n_estimators : int, default: 100
        The number of boosting stages to perform. Gradient boosting
        is fairly robust to over-fitting so a large number usually
        results in better performance.
        Values must be in the range `[1, inf)`.

    subsample : float, optional, default: 1.0
        The fraction of samples to be used for fitting the individual base
        learners. If smaller than 1.0 this results in Stochastic Gradient
        Boosting. `subsample` interacts with the parameter `n_estimators`.
        Choosing `subsample < 1.0` leads to a reduction of variance
        and an increase in bias.
        Values must be in the range `(0.0, 1.0]`.

    warm_start : bool, default: False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just erase the
        previous solution.

    dropout_rate : float, optional, default: 0.0
        If larger than zero, the residuals at each iteration are only computed
        from a random subset of base learners. The value corresponds to the
        percentage of base learners that are dropped. In each iteration,
        at least one base learner is dropped. This is an alternative regularization
        to shrinkage, i.e., setting `learning_rate < 1.0`.
        Values must be in the range `[0.0, 1.0)`.

    random_state : int seed, RandomState instance, or None, default: None
        The seed of the pseudo random number generator to use when
        shuffling the data.

    verbose : int, default: 0
        Enable verbose output. If 1 then it prints progress and performance
        once in a while.
        Values must be in the range `[0, inf)`.

    Attributes
    ----------
    coef_ : array, shape = (n_features + 1,)
        The aggregated coefficients. The first element `coef\_[0]` corresponds
        to the intercept. If loss is `coxph`, the intercept will always be zero.

    estimators_ : list of base learners
        The collection of fitted sub-estimators.

    train_score_ : ndarray, shape = (n_estimators,)
        The i-th score ``train_score_[i]`` is the loss of the
        model at iteration ``i`` on the in-bag sample.
        If ``subsample == 1`` this is the loss on the training data.

    oob_improvement_ : ndarray, shape = (n_estimators,)
        The improvement in loss on the out-of-bag samples
        relative to the previous iteration.
        ``oob_improvement_[0]`` is the improvement in
        loss of the first stage over the ``init`` estimator.
        Only available if ``subsample < 1.0``.

    oob_scores_ : ndarray of shape (n_estimators,)
        The full history of the loss values on the out-of-bag
        samples. Only available if ``subsample < 1.0``.

    oob_score_ : float
        The last value of the loss on the out-of-bag samples. It is
        the same as ``oob_scores_[-1]``. Only available if ``subsample < 1.0``.

    n_features_in_ : int
        Number of features seen during ``fit``.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during ``fit``. Defined only when `X`
        has feature names that are all strings.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    References
    ----------
    .. [1] Hothorn, T., Bühlmann, P., Dudoit, S., Molinaro, A., van der Laan, M. J.,
           "Survival ensembles", Biostatistics, 7(3), 355-73, 2006
    """

    _parameter_constraints = {
        "loss": [StrOptions(frozenset(LOSS_FUNCTIONS.keys()))],
        "learning_rate": [Interval(numbers.Real, 0.0, None, closed="left")],
        "n_estimators": [Interval(numbers.Integral, 1, None, closed="left")],
        "subsample": [Interval(numbers.Real, 0.0, 1.0, closed="right")],
        "warm_start": ["boolean"],
        "dropout_rate": [Interval(numbers.Real, 0.0, 1.0, closed="left")],
        "random_state": ["random_state"],
        "verbose": ["verbose"],
    }

    def __init__(
        self,
        *,
        loss="coxph",
        learning_rate=0.1,
        n_estimators=100,
        subsample=1.0,
        warm_start=False,
        dropout_rate=0,
        random_state=None,
        verbose=0,
    ):
        self.loss = loss
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.subsample = subsample
        self.warm_start = warm_start
        self.dropout_rate = dropout_rate
        self.random_state = random_state
        self.verbose = verbose

    @property
    def _predict_risk_score(self):
        return isinstance(self._loss, CoxPH)

    def _is_fitted(self):
        return len(getattr(self, "estimators_", [])) > 0

    def _init_state(self):
        self.estimators_ = np.empty(self.n_estimators, dtype=object)

        self.train_score_ = np.zeros(self.n_estimators, dtype=np.float64)
        # do oob?
        if self.subsample < 1.0:
            self.oob_improvement_ = np.zeros(
                self.n_estimators, dtype=np.float64)
            self.oob_scores_ = np.zeros(self.n_estimators, dtype=np.float64)
            self.oob_score_ = np.nan

        if self.dropout_rate > 0:
            self._scale = np.ones(int(self.n_estimators), dtype=float)

    def _resize_state(self):
        """Add additional ``n_estimators`` entries to all attributes."""
        # self.n_estimators is the number of additional est to fit
        total_n_estimators = self.n_estimators

        self.estimators_ = np.resize(self.estimators_, total_n_estimators)
        self.train_score_ = np.resize(self.train_score_, total_n_estimators)
        if self.subsample < 1 or hasattr(self, "oob_improvement_"):
            # if do oob resize arrays or create new if not available
            if hasattr(self, "oob_improvement_"):
                self.oob_improvement_ = np.resize(
                    self.oob_improvement_, total_n_estimators)
                self.oob_scores_ = np.resize(
                    self.oob_scores_, total_n_estimators)
                self.oob_score_ = np.nan
            else:
                self.oob_improvement_ = np.zeros(
                    total_n_estimators, dtype=np.float64)
                self.oob_scores_ = np.zeros(
                    (total_n_estimators,), dtype=np.float64)
                self.oob_score_ = np.nan

        if self.dropout_rate > 0:
            if not hasattr(self, "_scale"):
                raise ValueError(
                    "fitting with warm_start=True and dropout_rate > 0 is only "
                    "supported if the previous fit used dropout_rate > 0 too"
                )

            self._scale = np.resize(self._scale, total_n_estimators)
            self._scale[self.n_estimators_:] = 1

    def _clear_state(self):
        """Clear the state of the gradient boosting model."""
        if hasattr(self, "estimators_"):
            self.estimators_ = np.empty(0, dtype=object)
        if hasattr(self, "train_score_"):
            del self.train_score_
        if hasattr(self, "oob_improvement_"):
            del self.oob_improvement_
        if hasattr(self, "oob_scores_"):
            del self.oob_scores_
        if hasattr(self, "oob_score_"):
            del self.oob_score_
        if hasattr(self, "_rng"):
            del self._rng
        if hasattr(self, "_scale"):
            del self._scale

    def _update_with_dropout(self, i, X, raw_predictions, scale, random_state):
        # select base learners to be dropped for next iteration
        drop_model, n_dropped = _sample_binomial_plus_one(
            self.dropout_rate, i + 1, random_state)

        # adjust scaling factor of tree that is going to be trained in next iteration
        scale[i + 1] = 1.0 / (n_dropped + 1.0)

        raw_predictions[:] = 0
        for m in range(i + 1):
            if drop_model[m] == 1:
                # adjust scaling factor of dropped trees
                scale[m] *= n_dropped / (n_dropped + 1.0)
            else:
                # pseudoresponse of next iteration (without contribution of dropped trees)
                raw_predictions += self.learning_rate * \
                    scale[m] * self.estimators_[m].predict(X)

    def _fit(
        self,
        X,
        event,
        time,
        y_pred,
        sample_weight,
        random_state,
        begin_at_stage=0,
    ):  # noqa: C901
        n_samples = X.shape[0]
        # account for intercept
        y = np.fromiter(zip(event, time), dtype=[
                        ("event", bool), ("time", np.float64)])

        do_oob = self.subsample < 1.0
        if do_oob:
            n_inbag = max(1, int(self.subsample * n_samples))

        do_dropout = self.dropout_rate > 0
        if do_dropout:
            scale = self._scale

        if self.verbose:
            verbose_reporter = VerboseReporter(verbose=self.verbose)
            verbose_reporter.init(self, 0)

        # perform boosting iterations
        i = begin_at_stage
        for i in range(begin_at_stage, int(self.n_estimators)):
            # subsampling
            if do_oob:
                sample_mask = _random_sample_mask(
                    n_samples, n_inbag, random_state)
                subsample_weight = sample_weight * \
                    sample_mask.astype(np.float64)

                # OOB score before adding this stage
                y_oob_masked = y[~sample_mask]
                sample_weight_oob_masked = sample_weight[~sample_mask]
                if i == 0:  # store the initial loss to compute the OOB score
                    initial_loss = self._loss(
                        y_true=y_oob_masked,
                        raw_prediction=y_pred[~sample_mask],
                        sample_weight=sample_weight_oob_masked,
                    )
            else:
                subsample_weight = sample_weight

            residuals = self._loss.gradient(
                y, y_pred, sample_weight=sample_weight)

            best_learner = _fit_stage_componentwise(
                X, residuals, subsample_weight)
            self.estimators_[i] = best_learner

            if do_dropout and i < len(scale) - 1:
                self._update_with_dropout(i, X, y_pred, scale, random_state)
            else:
                y_pred += self.learning_rate * best_learner.predict(X)

            # track loss
            if do_oob:
                self.train_score_[i] = self._loss(
                    y_true=y[sample_mask],
                    raw_prediction=y_pred[sample_mask],
                    sample_weight=sample_weight[sample_mask],
                )
                self.oob_scores_[i] = self._loss(
                    y_true=y_oob_masked,
                    raw_prediction=y_pred[~sample_mask],
                    sample_weight=sample_weight_oob_masked,
                )
                previous_loss = initial_loss if i == 0 else self.oob_scores_[
                    i - 1]
                self.oob_improvement_[i] = previous_loss - self.oob_scores_[i]
                self.oob_score_ = self.oob_scores_[-1]
            else:
                # no need to fancy index w/ no subsampling
                self.train_score_[i] = self._loss(
                    y_true=y, raw_prediction=y_pred, sample_weight=sample_weight)

            if self.verbose > 0:
                verbose_reporter.update(i, self)

        return i + 1

    def fit(self, X, y, sample_weight=None):
        """Fit estimator.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix

        y : structured array, shape = (n_samples,)
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        sample_weight : array-like, shape = (n_samples,), optional
            Weights given to each sample. If omitted, all samples have weight 1.

        Returns
        -------
        self
        """
        self._validate_params()

        if not self.warm_start:
            self._clear_state()

        X = self._validate_data(X, ensure_min_samples=2)
        event, time = check_array_survival(X, y)

        sample_weight = _check_sample_weight(sample_weight, X)

        n_samples = X.shape[0]
        Xi = np.column_stack((np.ones(n_samples), X))

        self._loss = LOSS_FUNCTIONS[self.loss]()
        if isinstance(self._loss, (CensoredSquaredLoss, IPCWLeastSquaresError)):
            time = np.log(time)

        if not self._is_fitted():
            self._init_state()

            y_pred = np.zeros(n_samples, dtype=np.float64)

            begin_at_stage = 0

            self._rng = check_random_state(self.random_state)
        else:
            # add more estimators to fitted model
            # invariant: warm_start = True
            if self.n_estimators < self.estimators_.shape[0]:
                raise ValueError(
                    "n_estimators=%d must be larger or equal to "
                    "estimators_.shape[0]=%d when "
                    "warm_start==True" % (
                        self.n_estimators, self.estimators_.shape[0])
                )
            begin_at_stage = self.estimators_.shape[0]
            y_pred = self._raw_predict(Xi)
            self._resize_state()

            # apply dropout to last stage of previous fit
            if hasattr(self, "_scale") and self.dropout_rate > 0:
                # pylint: disable-next=access-member-before-definition
                self._update_with_dropout(
                    self.n_estimators_ - 1, Xi, y_pred, self._scale, self._rng)

        self.n_estimators_ = self._fit(
            Xi, event, time, y_pred, sample_weight, self._rng, begin_at_stage)

        self._set_baseline_model(X, event, time)
        return self

    def _set_baseline_model(self, X, event, time):
        if isinstance(self._loss, CoxPH):
            risk_scores = self._predict(X)
            self._baseline_model = BreslowEstimator().fit(risk_scores, event, time)
        else:
            self._baseline_model = None

    def _raw_predict(self, X):
        pred = np.zeros(X.shape[0], dtype=float)
        for estimator in self.estimators_:
            pred += self.learning_rate * estimator.predict(X)
        return pred

    def _predict(self, X):
        # account for intercept
        Xi = np.column_stack((np.ones(X.shape[0]), X))
        pred = self._raw_predict(Xi)
        return self._loss._scale_raw_prediction(pred)

    def predict(self, X):
        """Predict risk scores.

        If `loss='coxph'`, predictions can be interpreted as log hazard ratio
        corresponding to the linear predictor of a Cox proportional hazards
        model. If `loss='squared'` or `loss='ipcwls'`, predictions are the
        time to event.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix.

        Returns
        -------
        risk_score : array, shape = (n_samples,)
            Predicted risk scores.
        """
        check_is_fitted(self, "estimators_")
        X = self._validate_data(X, reset=False)

        return self._predict(X)

    def _get_baseline_model(self):
        if self._baseline_model is None:
            raise ValueError(
                "`fit` must be called with the loss option set to 'coxph'.")
        return self._baseline_model

    def predict_cumulative_hazard_function(self, X, return_array=False):
        """Predict cumulative hazard function.

        Only available if :meth:`fit` has been called with `loss = "coxph"`.

        The cumulative hazard function for an individual
        with feature vector :math:`x` is defined as

        .. math::

            H(t \\mid x) = \\exp(f(x)) H_0(t) ,

        where :math:`f(\\cdot)` is the additive ensemble of base learners,
        and :math:`H_0(t)` is the baseline hazard function,
        estimated by Breslow's estimator.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix.

        return_array : boolean, default: False
            If set, return an array with the cumulative hazard rate
            for each `self.unique_times_`, otherwise an array of
            :class:`survivalist.functions.StepFunction`.

        Returns
        -------
        cum_hazard : ndarray
            If `return_array` is set, an array with the cumulative hazard rate
            for each `self.unique_times_`, otherwise an array of length `n_samples`
            of :class:`survivalist.functions.StepFunction` instances will be returned.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from survivalist.datasets import load_whas500
        >>> from survivalist.ensemble import ComponentwiseGradientBoostingSurvivalAnalysis

        Load the data.

        >>> X, y = load_whas500()
        >>> X = X.astype(float)

        Fit the model.

        >>> estimator = ComponentwiseGradientBoostingSurvivalAnalysis(loss="coxph").fit(X, y)

        Estimate the cumulative hazard function for the first 10 samples.

        >>> chf_funcs = estimator.predict_cumulative_hazard_function(X.iloc[:10])

        Plot the estimated cumulative hazard functions.

        >>> for fn in chf_funcs:
        ...     plt.step(fn.x, fn(fn.x), where="post")
        ...
        >>> plt.ylim(0, 1)
        >>> plt.show()
        """
        return self._predict_cumulative_hazard_function(self._get_baseline_model(), self.predict(X), return_array)

    def predict_survival_function(self, X, return_array=False):
        """Predict survival function.

        Only available if :meth:`fit` has been called with `loss = "coxph"`.

        The survival function for an individual
        with feature vector :math:`x` is defined as

        .. math::

            S(t \\mid x) = S_0(t)^{\\exp(f(x)} ,

        where :math:`f(\\cdot)` is the additive ensemble of base learners,
        and :math:`S_0(t)` is the baseline survival function,
        estimated by Breslow's estimator.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix.

        return_array : boolean, default: False
            If set, return an array with the probability
            of survival for each `self.unique_times_`,
            otherwise an array of :class:`survivalist.functions.StepFunction`.

        Returns
        -------
        survival : ndarray
            If `return_array` is set, an array with the probability of
            survival for each `self.unique_times_`, otherwise an array of
            length `n_samples` of :class:`survivalist.functions.StepFunction`
            instances will be returned.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from survivalist.datasets import load_whas500
        >>> from survivalist.ensemble import ComponentwiseGradientBoostingSurvivalAnalysis

        Load the data.

        >>> X, y = load_whas500()
        >>> X = X.astype(float)

        Fit the model.

        >>> estimator = ComponentwiseGradientBoostingSurvivalAnalysis(loss="coxph").fit(X, y)

        Estimate the survival function for the first 10 samples.

        >>> surv_funcs = estimator.predict_survival_function(X.iloc[:10])

        Plot the estimated survival functions.

        >>> for fn in surv_funcs:
        ...     plt.step(fn.x, fn(fn.x), where="post")
        ...
        >>> plt.ylim(0, 1)
        >>> plt.show()
        """
        return self._predict_survival_function(self._get_baseline_model(), self.predict(X), return_array)

    @property
    def coef_(self):
        coef = np.zeros(self.n_features_in_ + 1, dtype=float)

        for estimator in self.estimators_:
            coef[estimator.component] += self.learning_rate * estimator.coef_

        return coef

    @property
    def unique_times_(self):
        return self._get_baseline_model().unique_times_

    @property
    def feature_importances_(self):
        imp = np.empty(self.n_features_in_ + 1, dtype=object)
        for i in range(imp.shape[0]):
            imp[i] = []

        for k, estimator in enumerate(self.estimators_):
            imp[estimator.component].append(k + 1)

        def _importance(x):
            if len(x) > 0:
                return np.min(x)
            return np.nan

        ret = np.array([_importance(x) for x in imp])
        return ret

    def _make_estimator(self, append=True, random_state=None):
        # we don't need _make_estimator
        raise NotImplementedError()


class GradientBoostingSurvivalAnalysis(BaseGradientBoosting, SurvivalAnalysisMixin):
    r"""Gradient-boosted Cox proportional hazard loss with
    regression trees as base learner.

    In each stage, a regression tree is fit on the negative gradient
    of the loss function.

    For more details on gradient boosting see [1]_ and [2]_. If `loss='coxph'`,
    the partial likelihood of the proportional hazards model is optimized as
    described in [3]_. If `loss='ipcwls'`, the accelerated failure time model with
    inverse-probability of censoring weighted least squares error is optimized as
    described in [4]_. When using a non-zero `dropout_rate`, regularization is
    applied during training following [5]_.

    See the :ref:`User Guide </user_guide/boosting.ipynb>` for examples.

    Parameters
    ----------
    loss : {'coxph', 'squared', 'ipcwls'}, optional, default: 'coxph'
        loss function to be optimized. 'coxph' refers to partial likelihood loss
        of Cox's proportional hazards model. The loss 'squared' minimizes a
        squared regression loss that ignores predictions beyond the time of censoring,
        and 'ipcwls' refers to inverse-probability of censoring weighted least squares error.

    learning_rate : float, optional, default: 0.1
        learning rate shrinks the contribution of each tree by `learning_rate`.
        There is a trade-off between `learning_rate` and `n_estimators`.
        Values must be in the range `[0.0, inf)`.

    n_estimators : int, default: 100
        The number of regression trees to create. Gradient boosting
        is fairly robust to over-fitting so a large number usually
        results in better performance.
        Values must be in the range `[1, inf)`.

    subsample : float, optional, default: 1.0
        The fraction of samples to be used for fitting the individual base
        learners. If smaller than 1.0 this results in Stochastic Gradient
        Boosting. `subsample` interacts with the parameter `n_estimators`.
        Choosing `subsample < 1.0` leads to a reduction of variance
        and an increase in bias.
        Values must be in the range `(0.0, 1.0]`.

    criterion : {'friedman_mse', 'squared_error'}, default: 'friedman_mse'
        The function to measure the quality of a split. Supported criteria are
        'friedman_mse' for the mean squared error with improvement score by
        Friedman, 'squared_error' for mean squared error. The default value of
        'friedman_mse' is generally the best as it can provide a better
        approximation in some cases.

    min_samples_split : int or float, optional, default: 2
        The minimum number of samples required to split an internal node:

        - If int, values must be in the range `[2, inf)`.
        - If float, values must be in the range `(0.0, 1.0]` and `min_samples_split`
          will be `ceil(min_samples_split * n_samples)`.

    min_samples_leaf : int or float, default: 1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, values must be in the range `[1, inf)`.
        - If float, values must be in the range `(0.0, 1.0)` and `min_samples_leaf`
          will be `ceil(min_samples_leaf * n_samples)`.

    min_weight_fraction_leaf : float, optional, default: 0.
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when `sample_weight` is not provided.
        Values must be in the range `[0.0, 0.5]`.

    max_depth : int or None, optional, default: 3
        Maximum depth of the individual regression estimators. The maximum
        depth limits the number of nodes in the tree. Tune this parameter
        for best performance; the best value depends on the interaction
        of the input variables. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        `min_samples_split` samples.
        If int, values must be in the range `[1, inf)`.

    min_impurity_decrease : float, optional, default: 0.
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

    random_state : int seed, RandomState instance, or None, default: None
        Controls the random seed given to each Tree estimator at each
        boosting iteration.
        In addition, it controls the random permutation of the features at
        each split.
        It also controls the random splitting of the training data to obtain a
        validation set if `n_iter_no_change` is not None.
        Pass an int for reproducible output across multiple function calls.

    max_features : int, float, string or None, optional, default: None
        The number of features to consider when looking for the best split:

        - If int, values must be in the range `[1, inf)`.
        - If float, values must be in the range `(0.0, 1.0]` and the features
          considered at each split will be `max(1, int(max_features * n_features_in_))`.
        - If 'sqrt', then `max_features=sqrt(n_features)`.
        - If 'log2', then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        Choosing `max_features < n_features` leads to a reduction of variance
        and an increase in bias.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int or None, optional, default: None
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        Values must be in the range `[2, inf)`.
        If `None`, then unlimited number of leaf nodes.

    warm_start : bool, default: False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just erase the
        previous solution.

    validation_fraction : float, default: 0.1
        The proportion of training data to set aside as validation set for
        early stopping. Values must be in the range `(0.0, 1.0)`.
        Only used if ``n_iter_no_change`` is set to an integer.

    n_iter_no_change : int, default: None
        ``n_iter_no_change`` is used to decide if early stopping will be used
        to terminate training when validation score is not improving. By
        default it is set to None to disable early stopping. If set to a
        number, it will set aside ``validation_fraction`` size of the training
        data as validation and terminate training when validation score is not
        improving in all of the previous ``n_iter_no_change`` numbers of
        iterations. The split is stratified.
        Values must be in the range `[1, inf)`.

    tol : float, default: 1e-4
        Tolerance for the early stopping. When the loss is not improving
        by at least tol for ``n_iter_no_change`` iterations (if set to a
        number), the training stops.
        Values must be in the range `[0.0, inf)`.

    dropout_rate : float, optional, default: 0.0
        If larger than zero, the residuals at each iteration are only computed
        from a random subset of base learners. The value corresponds to the
        percentage of base learners that are dropped. In each iteration,
        at least one base learner is dropped. This is an alternative regularization
        to shrinkage, i.e., setting `learning_rate < 1.0`.
        Values must be in the range `[0.0, 1.0)`.

    verbose : int, default: 0
        Enable verbose output. If 1 then it prints progress and performance
        once in a while (the more trees the lower the frequency). If greater
        than 1 then it prints progress and performance for every tree.
        Values must be in the range `[0, inf)`.

    ccp_alpha : non-negative float, optional, default: 0.0.
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed.
        Values must be in the range `[0.0, inf)`.

    Attributes
    ----------
    n_estimators_ : int
        The number of estimators as selected by early stopping (if
        ``n_iter_no_change`` is specified). Otherwise it is set to
        ``n_estimators``.

    feature_importances_ : ndarray, shape = (n_features,)
        The feature importances (the higher, the more important the feature).

    estimators_ : ndarray of DecisionTreeRegressor, shape = (n_estimators, 1)
        The collection of fitted sub-estimators.

    train_score_ : ndarray, shape = (n_estimators,)
        The i-th score ``train_score_[i]`` is the loss of the
        model at iteration ``i`` on the in-bag sample.
        If ``subsample == 1`` this is the loss on the training data.

    oob_improvement_ : ndarray, shape = (n_estimators,)
        The improvement in loss on the out-of-bag samples
        relative to the previous iteration.
        ``oob_improvement_[0]`` is the improvement in
        loss of the first stage over the ``init`` estimator.
        Only available if ``subsample < 1.0``.

    oob_scores_ : ndarray of shape (n_estimators,)
        The full history of the loss values on the out-of-bag
        samples. Only available if ``subsample < 1.0``.

    oob_score_ : float
        The last value of the loss on the out-of-bag samples. It is
        the same as ``oob_scores_[-1]``. Only available if ``subsample < 1.0``.

    n_features_in_ : int
        Number of features seen during ``fit``.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during ``fit``. Defined only when `X`
        has feature names that are all strings.

    max_features_ : int
        The inferred value of max_features.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    See also
    --------
    survivalist.ensemble.ComponentwiseGradientBoostingSurvivalAnalysis
        Gradient boosting with component-wise least squares as base learner.

    References
    ----------
    .. [1] J. H. Friedman, "Greedy function approximation: A gradient boosting machine,"
           The Annals of Statistics, 29(5), 1189–1232, 2001.
    .. [2] J. H. Friedman, "Stochastic gradient boosting,"
           Computational Statistics & Data Analysis, 38(4), 367–378, 2002.
    .. [3] G. Ridgeway, "The state of boosting,"
           Computing Science and Statistics, 172–181, 1999.
    .. [4] Hothorn, T., Bühlmann, P., Dudoit, S., Molinaro, A., van der Laan, M. J.,
           "Survival ensembles", Biostatistics, 7(3), 355-73, 2006.
    .. [5] K. V. Rashmi and R. Gilad-Bachrach,
           "DART: Dropouts meet multiple additive regression trees,"
           in 18th International Conference on Artificial Intelligence and Statistics,
           2015, 489–497.
    """

    _parameter_constraints = {
        **BaseGradientBoosting._parameter_constraints,
        "loss": [StrOptions(frozenset(LOSS_FUNCTIONS.keys()))],
        "dropout_rate": [Interval(numbers.Real, 0.0, 1.0, closed="left")],
    }

    def __init__(
        self,
        *,
        loss="coxph",
        learning_rate=0.1,
        n_estimators=100,
        subsample=1.0,
        criterion="friedman_mse",
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_depth=3,
        min_impurity_decrease=0.0,
        random_state=None,
        max_features=None,
        max_leaf_nodes=None,
        warm_start=False,
        validation_fraction=0.1,
        n_iter_no_change=None,
        tol=1e-4,
        dropout_rate=0.0,
        verbose=0,
        ccp_alpha=0.0,
    ):
        super().__init__(
            loss=loss,
            learning_rate=learning_rate,
            n_estimators=n_estimators,
            criterion=criterion,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_depth=max_depth,
            init="zero",
            subsample=subsample,
            max_features=max_features,
            random_state=random_state,
            verbose=verbose,
            max_leaf_nodes=max_leaf_nodes,
            warm_start=warm_start,
            min_impurity_decrease=min_impurity_decrease,
            validation_fraction=validation_fraction,
            n_iter_no_change=n_iter_no_change,
            tol=tol,
            ccp_alpha=ccp_alpha,
        )
        self.dropout_rate = dropout_rate

    def _encode_y(self, y, sample_weight):
        self.n_trees_per_iteration_ = 1
        return y

    def _get_loss(self, sample_weight):
        return LOSS_FUNCTIONS[self.loss]()

    @property
    def _predict_risk_score(self):
        return isinstance(self._loss, CoxPH)

    def _set_max_features(self):
        """Set self.max_features_."""
        if isinstance(self.max_features, str):
            if self.max_features == "sqrt":
                max_features = max(1, int(np.sqrt(self.n_features_in_)))
            elif self.max_features == "log2":
                max_features = max(1, int(np.log2(self.n_features_in_)))
        elif self.max_features is None:
            max_features = self.n_features_in_
        elif isinstance(self.max_features, (numbers.Integral, np.integer)):
            max_features = self.max_features
        else:  # float
            max_features = max(1, int(self.max_features * self.n_features_in_))

        self.max_features_ = max_features

    def _update_with_dropout(self, i, X, raw_predictions, k, scale, random_state):
        # select base learners to be dropped for next iteration
        drop_model, n_dropped = _sample_binomial_plus_one(
            self.dropout_rate, i + 1, random_state)

        # adjust scaling factor of tree that is going to be trained in next iteration
        scale[i + 1] = 1.0 / (n_dropped + 1.0)

        raw_predictions[:, k] = 0
        for m in range(i + 1):
            if drop_model[m] == 1:
                # adjust scaling factor of dropped trees
                scale[m] *= n_dropped / (n_dropped + 1.0)
            else:
                # pseudoresponse of next iteration (without contribution of dropped trees)
                raw_predictions[:, k] += self.learning_rate * \
                    scale[m] * self.estimators_[m, k].predict(X).ravel()

    def _fit_stage(
        self,
        i,
        X,
        y,
        raw_predictions,
        sample_weight,
        sample_mask,
        random_state,
        scale,
        X_csc=None,
        X_csr=None,
    ):
        """Fit another stage of ``n_classes_`` trees to the boosting model."""

        assert sample_mask.dtype == bool

        # whether to use dropout in next iteration
        do_dropout = self.dropout_rate > 0.0 and i < len(scale) - 1

        # Need to pass a copy of raw_predictions to negative_gradient()
        # because raw_predictions is partially updated at the end of the loop
        # in update_terminal_regions(), and gradients need to be evaluated at
        # iteration i - 1.
        raw_predictions_copy = raw_predictions.copy()

        neg_gradient = self._loss.gradient(
            y_true=y,
            raw_prediction=raw_predictions_copy,
            sample_weight=None,  # We pass sample_weights to the tree directly.
        )

        for k in range(self.n_trees_per_iteration_):
            # induce regression tree on the negative gradient
            tree = DecisionTreeRegressor(
                criterion=self.criterion,
                splitter="best",
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                min_samples_leaf=self.min_samples_leaf,
                min_weight_fraction_leaf=self.min_weight_fraction_leaf,
                min_impurity_decrease=self.min_impurity_decrease,
                max_features=self.max_features,
                max_leaf_nodes=self.max_leaf_nodes,
                random_state=random_state,
                ccp_alpha=self.ccp_alpha,
            )

            if self.subsample < 1.0:
                # no inplace multiplication!
                sample_weight = sample_weight * sample_mask.astype(np.float64)

            X = X_csc if X_csc is not None else X
            tree.fit(X, neg_gradient, sample_weight=sample_weight,
                     check_input=False)

            # add tree to ensemble
            self.estimators_[i, k] = tree

            # update tree leaves
            if do_dropout:
                self._update_with_dropout(
                    i, X, raw_predictions, k, scale, random_state)
            else:
                # update tree leaves
                X_for_tree_update = X_csr if X_csr is not None else X
                self._loss.update_terminal_regions(
                    tree.tree_,
                    X_for_tree_update,
                    y,
                    neg_gradient,
                    raw_predictions,
                    sample_weight,
                    sample_mask,
                    learning_rate=self.learning_rate,
                    k=k,
                )

        return raw_predictions

    def _fit_stages(  # noqa: C901
        self,
        X,
        y,
        raw_predictions,
        sample_weight,
        random_state,
        X_val,
        y_val,
        sample_weight_val,
        scale,
        begin_at_stage=0,
        monitor=None,
    ):
        """Iteratively fits the stages.

        For each stage it computes the progress (OOB, train score)
        and delegates to ``_fit_stage``.
        Returns the number of stages fit; might differ from ``n_estimators``
        due to early stopping.
        """
        n_samples = X.shape[0]
        do_oob = self.subsample < 1.0
        sample_mask = np.ones((n_samples,), dtype=bool)
        n_inbag = max(1, int(self.subsample * n_samples))

        if self.verbose:
            verbose_reporter = VerboseReporter(verbose=self.verbose)
            verbose_reporter.init(self, begin_at_stage)

        X_csc = csc_matrix(X) if issparse(X) else None
        X_csr = csr_matrix(X) if issparse(X) else None

        if self.n_iter_no_change is not None:
            loss_history = np.full(self.n_iter_no_change, np.inf)
            # We create a generator to get the predictions for X_val after
            # the addition of each successive stage
            y_val_pred_iter = self._staged_raw_predict(
                X_val, check_input=False)

        # perform boosting iterations
        i = begin_at_stage
        for i in range(begin_at_stage, self.n_estimators):
            # subsampling
            if do_oob:
                sample_mask = _random_sample_mask(
                    n_samples, n_inbag, random_state)
                # OOB score before adding this stage
                y_oob_masked = y[~sample_mask]
                sample_weight_oob_masked = sample_weight[~sample_mask]
                if i == 0:  # store the initial loss to compute the OOB score
                    initial_loss = self._loss(
                        y_true=y_oob_masked,
                        raw_prediction=raw_predictions[~sample_mask],
                        sample_weight=sample_weight_oob_masked,
                    )

            # fit next stage of trees
            raw_predictions = self._fit_stage(
                i,
                X,
                y,
                raw_predictions,
                sample_weight,
                sample_mask,
                random_state,
                scale,
                X_csc=X_csc,
                X_csr=X_csr,
            )

            # track loss
            if do_oob:
                self.train_score_[i] = self._loss(
                    y_true=y[sample_mask],
                    raw_prediction=raw_predictions[sample_mask],
                    sample_weight=sample_weight[sample_mask],
                )
                self.oob_scores_[i] = self._loss(
                    y_true=y_oob_masked,
                    raw_prediction=raw_predictions[~sample_mask],
                    sample_weight=sample_weight_oob_masked,
                )
                previous_loss = initial_loss if i == 0 else self.oob_scores_[
                    i - 1]
                self.oob_improvement_[i] = previous_loss - self.oob_scores_[i]
                self.oob_score_ = self.oob_scores_[-1]
            else:
                # no need to fancy index w/ no subsampling
                self.train_score_[i] = self._loss(
                    y_true=y,
                    raw_prediction=raw_predictions,
                    sample_weight=sample_weight,
                )

            if self.verbose > 0:
                verbose_reporter.update(i, self)

            if monitor is not None:
                early_stopping = monitor(i, self, locals())
                if early_stopping:
                    break

            # We also provide an early stopping based on the score from
            # validation set (X_val, y_val), if n_iter_no_change is set
            if self.n_iter_no_change is not None:
                # By calling next(y_val_pred_iter), we get the predictions
                # for X_val after the addition of the current stage
                validation_loss = self._loss(y_val, next(
                    y_val_pred_iter), sample_weight_val)

                # Require validation_score to be better (less) than at least
                # one of the last n_iter_no_change evaluations
                if np.any(validation_loss + self.tol < loss_history):
                    loss_history[i % len(loss_history)] = validation_loss
                else:
                    break

        return i + 1

    def _init_state(self):
        super()._init_state()

        if self.dropout_rate > 0.0:
            self._scale = np.ones(self.n_estimators, dtype=float)

    def _resize_state(self):
        super()._resize_state()

        if self.dropout_rate > 0:
            if not hasattr(self, "_scale"):
                raise ValueError(
                    "fitting with warm_start=True and dropout_rate > 0 is only "
                    "supported if the previous fit used dropout_rate > 0 too"
                )

            self._scale = np.resize(self._scale, self.n_estimators)
            self._scale[self.n_estimators_:] = 1

    def _shrink_state(self, n_stages):
        self.estimators_ = self.estimators_[:n_stages]
        self.train_score_ = self.train_score_[:n_stages]
        if hasattr(self, "oob_improvement_"):
            self.oob_improvement_ = self.oob_improvement_[:n_stages]
            self.oob_scores_ = self.oob_scores_[:n_stages]
            self.oob_score_ = self.oob_scores_[-1]
        if hasattr(self, "_scale"):
            self._scale = self._scale[:n_stages]

    def fit(self, X, y, sample_weight=None, monitor=None):
        """Fit the gradient boosting model.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix

        y : structured array, shape = (n_samples,)
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        sample_weight : array-like, shape = (n_samples,), optional
            Weights given to each sample. If omitted, all samples have weight 1.

        monitor : callable, optional
            The monitor is called after each iteration with the current
            iteration, a reference to the estimator and the local variables of
            ``_fit_stages`` as keyword arguments ``callable(i, self,
            locals())``. If the callable returns ``True`` the fitting procedure
            is stopped. The monitor can be used for various things such as
            computing held-out estimates, early stopping, model introspect, and
            snapshoting.

        Returns
        -------
        self : object
            Returns self.
        """
        self._validate_params()

        if not self.warm_start:
            self._clear_state()

        X = self._validate_data(
            X,
            ensure_min_samples=2,
            order="C",
            accept_sparse=["csr", "csc", "coo"],
            dtype=DTYPE,
        )
        event, time = check_array_survival(X, y)

        sample_weight_is_none = sample_weight is None
        sample_weight = _check_sample_weight(sample_weight, X)

        if sample_weight_is_none:
            y = self._encode_y(y=y, sample_weight=None)
        else:
            y = self._encode_y(y=y, sample_weight=sample_weight)

        self._set_max_features()

        # self.loss is guaranteed to be a string
        self._loss = self._get_loss(sample_weight=sample_weight)

        if isinstance(self._loss, (CensoredSquaredLoss, IPCWLeastSquaresError)):
            time = np.log(time)

        if self.n_iter_no_change is not None:
            (
                X_train,
                X_val,
                event_train,
                event_val,
                time_train,
                time_val,
                sample_weight_train,
                sample_weight_val,
            ) = train_test_split(
                X,
                event,
                time,
                sample_weight,
                random_state=self.random_state,
                test_size=self.validation_fraction,
                stratify=event,
            )
            y_val = np.fromiter(
                zip(event_val, time_val),
                dtype=[("event", bool), ("time", np.float64)],
            )
        else:
            X_train, sample_weight_train = X, sample_weight
            event_train, time_train = event, time
            X_val = y_val = sample_weight_val = None

        y_train = np.fromiter(
            zip(event_train, time_train),
            dtype=[("event", bool), ("time", np.float64)],
        )
        n_samples = X_train.shape[0]

        # First time calling fit.
        if not self._is_fitted():
            # init state
            self._init_state()

            raw_predictions = np.zeros(
                shape=(n_samples, self.n_trees_per_iteration_),
                dtype=np.float64,
            )

            begin_at_stage = 0

            # The rng state must be preserved if warm_start is True
            self._rng = check_random_state(self.random_state)

        # warm start: this is not the first time fit was called
        else:
            # add more estimators to fitted model
            # invariant: warm_start = True
            if self.n_estimators < self.estimators_.shape[0]:
                raise ValueError(
                    "n_estimators=%d must be larger or equal to "
                    "estimators_.shape[0]=%d when "
                    "warm_start==True" % (
                        self.n_estimators, self.estimators_.shape[0])
                )
            begin_at_stage = self.estimators_.shape[0]
            # The requirements of _raw_predict
            # are more constrained than fit. It accepts only CSR
            # matrices. Finite values have already been checked in _validate_data.
            X_train = check_array(
                X_train,
                dtype=DTYPE,
                order="C",
                accept_sparse="csr",
                force_all_finite=False,
            )
            raw_predictions = self._raw_predict(X_train)
            self._resize_state()

            # apply dropout to last stage of previous fit
            if hasattr(self, "_scale") and self.dropout_rate > 0:
                for k in range(self.n_trees_per_iteration_):
                    self._update_with_dropout(
                        # pylint: disable-next=access-member-before-definition
                        self.n_estimators_ - 1,
                        X_train,
                        raw_predictions,
                        k,
                        self._scale,
                        self._rng,
                    )

        scale = getattr(self, "_scale", None)

        # fit the boosting stages
        n_stages = self._fit_stages(
            X_train,
            y_train,
            raw_predictions,
            sample_weight_train,
            self._rng,
            X_val,
            y_val,
            sample_weight_val,
            scale,
            begin_at_stage,
            monitor,
        )
        # change shape of arrays after fit (early-stopping or additional tests)
        if n_stages != self.estimators_.shape[0]:
            self._shrink_state(n_stages)
        self.n_estimators_ = n_stages

        self._set_baseline_model(X_train, event_train, time_train)

        return self

    def _set_baseline_model(self, X, event, time):
        if isinstance(self._loss, CoxPH):
            X_pred = X
            if issparse(X):
                X_pred = X.asformat("csr")
            risk_scores = self._predict(X_pred)
            self._baseline_model = BreslowEstimator().fit(risk_scores, event, time)
        else:
            self._baseline_model = None

    def _dropout_predict_stage(self, X, i, K, score):
        for k in range(K):
            tree = self.estimators_[i, k].tree_
            score += self.learning_rate * \
                self._scale[i] * tree.predict(X).reshape((-1, 1))
        return score

    def _dropout_raw_predict(self, X):
        raw_predictions = self._raw_predict_init(X)

        n_estimators, K = self.estimators_.shape
        for i in range(n_estimators):
            self._dropout_predict_stage(X, i, K, raw_predictions)

        return raw_predictions

    def _dropout_staged_raw_predict(self, X):
        X = self._validate_data(X, dtype=DTYPE, order="C", accept_sparse="csr")
        raw_predictions = self._raw_predict_init(X)

        n_estimators, K = self.estimators_.shape
        for i in range(n_estimators):
            self._dropout_predict_stage(X, i, K, raw_predictions)
            yield raw_predictions.copy()

    def _raw_predict(self, X):
        # if dropout wasn't used during training, proceed as usual,
        # otherwise consider scaling factor of individual trees
        if not hasattr(self, "_scale"):
            return super()._raw_predict(X)
        return self._dropout_raw_predict(X)

    def _init_decision_function(self, X):  # pragma: no cover
        return super()._init_decision_function(X).reshape(-1, 1)

    def _decision_function(self, X):  # pragma: no cover
        return self._raw_predict(X)

    def _predict(self, X):
        score = self._raw_predict(X)
        if score.shape[1] == 1:
            score = score.ravel()
        return self._loss._scale_raw_prediction(score)

    def predict(self, X):
        """Predict risk scores.

        If `loss='coxph'`, predictions can be interpreted as log hazard ratio
        similar to the linear predictor of a Cox proportional hazards
        model. If `loss='squared'` or `loss='ipcwls'`, predictions are the
        time to event.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape = (n_samples,)
            The risk scores.
        """
        check_is_fitted(self, "estimators_")

        X = self._validate_data(X, reset=False, order="C",
                                accept_sparse="csr", dtype=DTYPE)
        return self._predict(X)

    def staged_predict(self, X):
        """Predict risk scores at each stage for X.

        This method allows monitoring (i.e. determine error on testing set)
        after each stage.

        If `loss='coxph'`, predictions can be interpreted as log hazard ratio
        similar to the linear predictor of a Cox proportional hazards
        model. If `loss='squared'` or `loss='ipcwls'`, predictions are the
        time to event.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : generator of array of shape = (n_samples,)
            The predicted value of the input samples.
        """
        check_is_fitted(self, "estimators_")

        # if dropout wasn't used during training, proceed as usual,
        # otherwise consider scaling factor of individual trees
        if not hasattr(self, "_scale"):
            predictions_iter = self._staged_raw_predict(X)
        else:
            predictions_iter = self._dropout_staged_raw_predict(X)

        for raw_predictions in predictions_iter:
            y = self._loss._scale_raw_prediction(raw_predictions)
            yield y.ravel()

    def _get_baseline_model(self):
        if self._baseline_model is None:
            raise ValueError(
                "`fit` must be called with the loss option set to 'coxph'.")
        return self._baseline_model

    def predict_cumulative_hazard_function(self, X, return_array=False):
        """Predict cumulative hazard function.

        Only available if :meth:`fit` has been called with `loss = "coxph"`.

        The cumulative hazard function for an individual
        with feature vector :math:`x` is defined as

        .. math::

            H(t \\mid x) = \\exp(f(x)) H_0(t) ,

        where :math:`f(\\cdot)` is the additive ensemble of base learners,
        and :math:`H_0(t)` is the baseline hazard function,
        estimated by Breslow's estimator.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix.

        return_array : boolean, default: False
            If set, return an array with the cumulative hazard rate
            for each `self.unique_times_`, otherwise an array of
            :class:`survivalist.functions.StepFunction`.

        Returns
        -------
        cum_hazard : ndarray
            If `return_array` is set, an array with the cumulative hazard rate
            for each `self.unique_times_`, otherwise an array of length `n_samples`
            of :class:`survivalist.functions.StepFunction` instances will be returned.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from survivalist.datasets import load_whas500
        >>> from survivalist.ensemble import GradientBoostingSurvivalAnalysis

        Load the data.

        >>> X, y = load_whas500()
        >>> X = X.astype(float)

        Fit the model.

        >>> estimator = GradientBoostingSurvivalAnalysis(loss="coxph").fit(X, y)

        Estimate the cumulative hazard function for the first 10 samples.

        >>> chf_funcs = estimator.predict_cumulative_hazard_function(X.iloc[:10])

        Plot the estimated cumulative hazard functions.

        >>> for fn in chf_funcs:
        ...     plt.step(fn.x, fn(fn.x), where="post")
        ...
        >>> plt.ylim(0, 1)
        >>> plt.show()
        """
        return self._predict_cumulative_hazard_function(self._get_baseline_model(), self.predict(X), return_array)

    def predict_survival_function(self, X, return_array=False):
        """Predict survival function.

        Only available if :meth:`fit` has been called with `loss = "coxph"`.

        The survival function for an individual
        with feature vector :math:`x` is defined as

        .. math::

            S(t \\mid x) = S_0(t)^{\\exp(f(x)} ,

        where :math:`f(\\cdot)` is the additive ensemble of base learners,
        and :math:`S_0(t)` is the baseline survival function,
        estimated by Breslow's estimator.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix.

        return_array : boolean, default: False
            If set, return an array with the probability
            of survival for each `self.unique_times_`,
            otherwise an array of :class:`survivalist.functions.StepFunction`.

        Returns
        -------
        survival : ndarray
            If `return_array` is set, an array with the probability of
            survival for each `self.unique_times_`, otherwise an array of
            length `n_samples` of :class:`survivalist.functions.StepFunction`
            instances will be returned.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from survivalist.datasets import load_whas500
        >>> from survivalist.ensemble import GradientBoostingSurvivalAnalysis

        Load the data.

        >>> X, y = load_whas500()
        >>> X = X.astype(float)

        Fit the model.

        >>> estimator = GradientBoostingSurvivalAnalysis(loss="coxph").fit(X, y)

        Estimate the survival function for the first 10 samples.

        >>> surv_funcs = estimator.predict_survival_function(X.iloc[:10])

        Plot the estimated survival functions.

        >>> for fn in surv_funcs:
        ...     plt.step(fn.x, fn(fn.x), where="post")
        ...
        >>> plt.ylim(0, 1)
        >>> plt.show()
        """
        return self._predict_survival_function(self._get_baseline_model(), self.predict(X), return_array)

    @property
    def unique_times_(self):
        return self._get_baseline_model().unique_times_



from collections import OrderedDict
from contextlib import nullcontext as does_not_raise

import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd
import pandas.testing as tm
import pytest

from survivalist.testing import FixtureParameterFactory
from survivalist.util import Surv, _PropertyAvailableIfDescriptor, property_available_if, safe_concat


class ConcatCasesFactory(FixtureParameterFactory):
    @property
    def rnd(self):
        return np.random.RandomState(14)

    def to_data_frame(self, data):
        return pd.DataFrame.from_dict(OrderedDict(data))

    def make_numeric_series(self, name):
        return pd.Series(self.rnd.randn(100), name=name)

    def make_categorical_5_series(self, name):
        return pd.Series(
            pd.Categorical.from_codes(self.rnd.binomial(4, 0.6, 100), ["C1", "C2", "C3", "C4", "C5"]), name=name
        )

    def make_categorical_4_series(self, name):
        return pd.Series(pd.Categorical.from_codes(self.rnd.binomial(3, 0.6, 100), ["C1", "C2", "C3", "C4"]), name=name)

    def make_categorical_3_series(self, name):
        return pd.Series(pd.Categorical.from_codes(self.rnd.binomial(2, 0.6, 100), ["C1", "C2", "C3"]), name=name)

    def make_binary_series(self, name):
        return pd.Series(pd.Categorical.from_codes(self.rnd.binomial(1, 0.6, 100), ["Yes", "No"]), name=name)


class ConcatCasesAxes1(ConcatCasesFactory):
    def data_numeric(self):
        a = self.make_numeric_series("col_A")
        b = self.make_numeric_series("col_B")

        expected = [(a.name, a), (b.name, b)]
        return (a, b), self.to_data_frame(expected), does_not_raise()

    def data_numeric_categorical(self):
        a = self.make_numeric_series("col_A")
        b = self.make_categorical_5_series("col_B")

        expected = [(a.name, a), (b.name, b)]
        return (a, b), self.to_data_frame(expected), does_not_raise()

    def data_frame_numeric_categorical(self):
        numeric_df = self.to_data_frame(
            [
                ("col_A", self.make_numeric_series("col_A")),
                ("col_B", self.make_categorical_5_series("col_B")),
            ]
        )

        cat_series = self.make_categorical_5_series("col_C")

        expected_df = numeric_df.copy()
        expected_df.loc[:, "col_C"] = cat_series
        return (numeric_df, cat_series), expected_df, does_not_raise()

    def data_duplicate_columns(self):
        numeric_df = self.to_data_frame(
            [
                ("col_N", self.make_numeric_series("col_N")),
                ("col_B", self.make_categorical_5_series("col_B")),
                ("col_A", self.make_categorical_5_series("col_A")),
            ]
        )

        cat_df = pd.DataFrame.from_dict(
            OrderedDict(
                [
                    ("col_A", self.make_categorical_5_series("col_A")),
                    ("col_C", self.make_binary_series("col_C")),
                ]
            )
        )

        err = pytest.raises(ValueError, match="duplicate columns col_A")
        return (numeric_df, cat_df), None, err


class ConcatCasesAxes0(ConcatCasesFactory):
    def data_categorical(self):
        a = self.to_data_frame(
            [
                ("col_A", self.make_categorical_3_series("col_A")),
                ("col_B", self.make_numeric_series("col_B")),
            ]
        )
        b = self.to_data_frame(
            [("col_A", self.make_categorical_3_series("col_A")), ("col_B", self.make_numeric_series("col_B"))]
        )

        expected = [
            (
                "col_A",
                pd.Series(
                    pd.Categorical.from_codes(
                        np.r_[a.col_A.cat.codes.values, b.col_A.cat.codes.values], ["C1", "C2", "C3"]
                    )
                ),
            ),
            ("col_B", np.r_[a.col_B.values, b.col_B.values]),
        ]
        expected_df = self.to_data_frame(expected)
        expected_df.index = pd.Index(a.index.tolist() + b.index.tolist())

        return (a, b), expected_df, does_not_raise()

    def data_categorical_mismatch(self):
        a = self.to_data_frame(
            [
                ("col_A", self.make_categorical_3_series("col_A")),
                ("col_B", self.make_numeric_series("col_B")),
            ]
        )
        b = self.to_data_frame(
            [
                ("col_A", self.make_categorical_4_series("col_A")),
                ("col_B", self.make_numeric_series("col_B")),
            ]
        )

        err = pytest.raises(ValueError, match="categories for column col_A do not match")
        return (a, b), None, err


@pytest.mark.parametrize("inputs,expected_df,expected_error", ConcatCasesAxes1().get_cases())
def test_concat_axis_1(inputs, expected_df, expected_error):
    with expected_error:
        actual_df = safe_concat(inputs, axis=1)

    if expected_df is not None:
        tm.assert_frame_equal(actual_df, expected_df)


@pytest.mark.parametrize("inputs,expected_df,expected_error", ConcatCasesAxes0().get_cases())
def test_concat_axis_0(inputs, expected_df, expected_error):
    with expected_error:
        actual_df = safe_concat(inputs, axis=0)

    if expected_df is not None:
        tm.assert_frame_equal(actual_df, expected_df)


class SurvCases(FixtureParameterFactory):
    @property
    def event_and_time(self):
        event = np.random.binomial(1, 0.5, size=100)
        time = np.exp(np.random.randn(100))
        return event, time


class SurvArrayCases(SurvCases):
    def get_surv_arrays(self, event_name="event", time_name="time"):
        event, time = self.event_and_time

        expected = np.empty(dtype=[(event_name, bool), (time_name, float)], shape=event.shape[0])
        expected[event_name] = event.astype(bool)
        expected[time_name] = time

        return (event, time), expected

    def data_list(self):
        (event, time), expected = self.get_surv_arrays()

        inputs = (list(event.astype(bool)), list(time))
        return inputs, {}, expected, does_not_raise()

    def data_bool(self):
        (event, time), expected = self.get_surv_arrays()

        inputs = (event.astype(bool), time)
        return inputs, {}, expected, does_not_raise()

    def data_with_names(self):
        (event, time), expected = self.get_surv_arrays("death", "survival_time")

        inputs = (event.astype(bool), time)
        kwargs = {"name_time": "survival_time", "name_event": "death"}
        return inputs, kwargs, expected, does_not_raise()

    def data_with_one_name_1(self):
        (event, time), expected = self.get_surv_arrays("death")

        inputs = (event.astype(bool), time)
        kwargs = {"name_event": "death"}
        return inputs, kwargs, expected, does_not_raise()

    def data_with_one_name_2(self):
        (event, time), expected = self.get_surv_arrays("event", "survival_time")

        inputs = (event.astype(bool), time)
        kwargs = {"name_time": "survival_time"}
        return inputs, kwargs, expected, does_not_raise()

    def data_array_int_event(self):
        inputs, expected = self.get_surv_arrays()

        return inputs, {}, expected, does_not_raise()

    def data_int_time(self):
        event, time = self.event_and_time
        time += 1
        time *= time

        expected = np.empty(dtype=[("event", bool), ("time", float)], shape=event.shape[0])
        expected["event"] = event.astype(bool)
        expected["time"] = time.astype(int)

        inputs = (event.astype(bool), time.astype(int))
        return inputs, {}, expected, does_not_raise()

    def data_float(self):
        (event, time), expected = self.get_surv_arrays()

        inputs = (event.astype(float), time)
        return inputs, {}, expected, does_not_raise()

    def data_shape_mismatch_0(self):
        event, time = self.event_and_time

        msg = "Found input variables with inconsistent numbers of samples"
        err = pytest.raises(ValueError, match=msg)
        inputs = (event[1:], time)
        return inputs, {}, None, err

    def data_shape_mismatch_1(self):
        event, time = self.event_and_time
        msg = "Found input variables with inconsistent numbers of samples"
        err = pytest.raises(ValueError, match=msg)
        inputs = (event, time[1:])
        return inputs, {}, None, err

    def data_event_value_wrong_1(self):
        event, time = self.event_and_time
        event += 1

        err = pytest.raises(ValueError, match="non-boolean event indicator must contain 0 and 1 only")
        return (event, time), {}, None, err

    def data_event_value_wrong_2(self):
        event, time = self.event_and_time
        event -= 1

        err = pytest.raises(ValueError, match="non-boolean event indicator must contain 0 and 1 only")
        return (event, time), {}, None, err

    def data_event_value_wrong_3(self):
        event, time = self.event_and_time
        event[event == 0] = 3

        err = pytest.raises(ValueError, match="non-boolean event indicator must contain 0 and 1 only")

        return (event, time), {}, None, err

    def data_event_value_wrong_4(self):
        event, time = self.event_and_time
        event[1] = 3

        err = pytest.raises(ValueError, match="event indicator must be binary")
        return (event, time), {}, None, err

    def data_event_value_wrong_5(self):
        event, time = self.event_and_time
        event = np.arange(event.shape[0])

        err = pytest.raises(ValueError, match="event indicator must be binary")
        return (event, time), {}, None, err

    def data_names_match(self):
        event, time = self.event_and_time

        err = pytest.raises(ValueError, match="name_time must be different from name_event")
        kwargs = {"name_event": "time_and_event", "name_time": "time_and_event"}
        return (event, time), kwargs, None, err


@pytest.mark.parametrize("args,kwargs,expected,expected_error", SurvArrayCases().get_cases())
def test_from_arrays(args, kwargs, expected, expected_error):
    with expected_error:
        y = Surv.from_arrays(*args, **kwargs)

    if expected is not None:
        assert_array_equal(y, expected)


class SurvDataFrameCases(SurvCases):
    def get_surv_data_frame(self, event_name="event", time_name="time"):
        event, time = self.event_and_time
        df = pd.DataFrame({event_name: event, time_name: time})

        expected = np.empty(dtype=[(str(event_name), bool), (str(time_name), float)], shape=100)
        expected[str(event_name)] = event.astype(bool)
        expected[str(time_name)] = time

        return df, expected

    def data_bool(self):
        data, expected = self.get_surv_data_frame()
        data["event"] = data["event"].astype(bool)

        inputs = ("event", "time", data)
        return inputs, expected, does_not_raise()

    def data_int(self):
        data, expected = self.get_surv_data_frame()
        inputs = ("event", "time", data)
        return inputs, expected, does_not_raise()

    def data_float(self):
        data, expected = self.get_surv_data_frame()
        data["event"] = data["event"].astype(float)
        inputs = ("event", "time", data)
        return inputs, expected, does_not_raise()

    def data_no_str_columns(self):
        data, expected = self.get_surv_data_frame(event_name=0, time_name=1)
        inputs = (0, 1, data)
        return inputs, expected, does_not_raise()

    def data_column_names(self):
        data, expected = self.get_surv_data_frame(event_name="death", time_name="time_to_death")
        inputs = ("death", "time_to_death", data)
        return inputs, expected, does_not_raise()

    def data_no_such_column_0(self):
        data, _ = self.get_surv_data_frame()

        err = pytest.raises(KeyError, match="unknown")
        inputs = ("unknown", "time", data)
        return inputs, None, err

    def data_no_such_column_1(self):
        data, _ = self.get_surv_data_frame()

        err = pytest.raises(KeyError, match="unknown")
        inputs = ("event", "unknown", data)
        return inputs, None, err

    def data_wrong_class_0(self):
        data, _ = self.get_surv_data_frame()

        err = pytest.raises(TypeError, match=r"expected pandas.DataFrame, but got <class 'dict'>")
        inputs = ("event", "time", data.to_dict())
        return inputs, None, err

    def data_wrong_class_1(self):
        data, _ = self.get_surv_data_frame()

        err = pytest.raises(TypeError, match=r"expected pandas.DataFrame, but got <class 'numpy.ndarray'>")
        inputs = ("event", "time", data.values)
        return inputs, None, err


@pytest.mark.parametrize("args,expected,expected_error", SurvDataFrameCases().get_cases())
def test_from_dataframe(args, expected, expected_error):
    with expected_error:
        y = Surv.from_dataframe(*args)

    if expected is not None:
        assert_array_equal(y, expected)


def test_cond_avail_property():
    class WithCondProp:
        def __init__(self, val):
            self.avail = False
            self._prop = val

        @property_available_if(lambda self: self.avail)
        def prop(self):
            return self._prop

        no_prop = _PropertyAvailableIfDescriptor(lambda self: self.avail, fget=None)

    testval = 43
    msg = "has no attribute 'prop'"

    assert WithCondProp.prop is not None

    test_obj = WithCondProp(testval)
    with pytest.raises(AttributeError, match=msg):
        _ = test_obj.prop
    assert test_obj.avail is False

    test_obj.avail = True
    assert test_obj.prop == testval

    test_obj.avail = False
    with pytest.raises(AttributeError, match=msg):
        _ = test_obj.prop

    test_obj.avail = True
    with pytest.raises(AttributeError, match="has no getter"):
        _ = test_obj.no_prop



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numbers
import warnings

import numpy as np
from scipy.linalg import solve
from sklearn.base import BaseEstimator
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import check_array, check_is_fitted

from ..base import SurvivalAnalysisMixin
from ..functions import StepFunction
from ..nonparametric import _compute_counts
from ..util import check_array_survival

__all__ = ["CoxPHSurvivalAnalysis", "BreslowEstimator"]


class BreslowEstimator:
    """Breslow's estimator of the cumulative hazard function.

    Attributes
    ----------
    cum_baseline_hazard_ : :class:`sksurv.functions.StepFunction`
        Cumulative baseline hazard function.

    baseline_survival_ : :class:`sksurv.functions.StepFunction`
        Baseline survival function.

    unique_times_ : ndarray
        Unique event times.
    """

    def fit(self, linear_predictor, event, time):
        """Compute baseline cumulative hazard function.

        Parameters
        ----------
        linear_predictor : array-like, shape = (n_samples,)
            Linear predictor of risk: `X @ coef`.

        event : array-like, shape = (n_samples,)
            Contains binary event indicators.

        time : array-like, shape = (n_samples,)
            Contains event/censoring times.

        Returns
        -------
        self
        """
        # Stabilize risk scores (if needed)
        max_val = np.max(linear_predictor)
        risk_score = np.exp(linear_predictor - max_val)  # Ratios preserved
        order = np.argsort(time, kind="mergesort")
        risk_score = risk_score[order]
        uniq_times, n_events, n_at_risk, _ = _compute_counts(
            event, time, order)

        divisor = np.empty(n_at_risk.shape, dtype=float)
        value = np.sum(risk_score)
        divisor[0] = value
        k = 0
        for i in range(1, len(n_at_risk)):
            d = n_at_risk[i - 1] - n_at_risk[i]
            value -= risk_score[k: (k + d)].sum()
            k += d
            divisor[i] = value

        assert k == n_at_risk[0] - n_at_risk[-1]

        y = np.cumsum(n_events / (divisor + np.finfo(float).eps))
        self.cum_baseline_hazard_ = StepFunction(uniq_times, y)
        # Apply log-sum-exp trick: log(exp(-y)) = -y
        log_survival = -y
        # Clip to avoid overflow in exp
        log_survival = np.clip(log_survival, -700, 700)
        np_exp_y = np.exp(log_survival)
        self.baseline_survival_ = StepFunction(uniq_times, np_exp_y)
        self.unique_times_ = uniq_times
        return self

    def get_cumulative_hazard_function(self, linear_predictor):
        """Predict cumulative hazard function.

        Parameters
        ----------
        linear_predictor : array-like, shape = (n_samples,)
            Linear predictor of risk: `X @ coef`.

        Returns
        -------
        cum_hazard : ndarray, shape = (n_samples,)
            Predicted cumulative hazard functions.
        """
        risk_score = np.exp(linear_predictor)
        n_samples = risk_score.shape[0]
        funcs = np.empty(n_samples, dtype=object)
        for i in range(n_samples):
            funcs[i] = StepFunction(
                x=self.cum_baseline_hazard_.x,
                y=self.cum_baseline_hazard_.y,
                a=risk_score[i],
            )
        return funcs

    def get_survival_function(self, linear_predictor):
        """Predict survival function.

        Parameters
        ----------
        linear_predictor : array-like, shape = (n_samples,)
            Linear predictor of risk: `X @ coef`.

        Returns
        -------
        survival : ndarray, shape = (n_samples,)
            Predicted survival functions.
        """
        risk_score = np.exp(linear_predictor)
        n_samples = risk_score.shape[0]
        funcs = np.empty(n_samples, dtype=object)
        for i in range(n_samples):
            funcs[i] = StepFunction(
                x=self.baseline_survival_.x,
                y=np.power(self.baseline_survival_.y, risk_score[i]),
            )
        return funcs


class CoxPHOptimizer:
    """Negative partial log-likelihood of Cox proportional hazards model"""

    def __init__(self, X, event, time, alpha, ties):
        # sort descending
        o = np.argsort(-time, kind="mergesort")
        self.x = X[o, :]
        self.event = event[o]
        self.time = time[o]
        self.alpha = alpha
        self.no_alpha = np.all(self.alpha < np.finfo(self.alpha.dtype).eps)
        self._is_breslow = ties == "breslow"

    def nlog_likelihood(self, w):
        """Compute negative partial log-likelihood

        Parameters
        ----------
        w : array, shape = (n_features,)
            Estimate of coefficients

        Returns
        -------
        loss : float
            Average negative partial log-likelihood
        """
        time = self.time
        n_samples = self.x.shape[0]
        breslow = self._is_breslow
        xw = np.dot(self.x, w)

        loss = 0
        risk_set = 0
        k = 0
        while k < n_samples:
            ti = time[k]
            numerator = 0
            n_events = 0
            risk_set2 = 0
            while k < n_samples and ti == time[k]:
                if self.event[k]:
                    numerator += xw[k]
                    risk_set2 += np.exp(xw[k])
                    n_events += 1
                else:
                    risk_set += np.exp(xw[k])
                k += 1

            if n_events > 0:
                if breslow:
                    risk_set += risk_set2
                    loss -= (numerator - n_events *
                             np.log(risk_set)) / n_samples
                else:
                    numerator /= n_events
                    for _ in range(n_events):
                        risk_set += risk_set2 / n_events
                        loss -= (numerator - np.log(risk_set)) / n_samples

        # add regularization term to log-likelihood
        return loss + np.sum(self.alpha * np.square(w)) / (2.0 * n_samples)

    def update(self, w, offset=0):
        """Compute gradient and Hessian matrix with respect to `w`."""
        time = self.time
        x = self.x
        breslow = self._is_breslow
        exp_xw = np.exp(offset + np.dot(x, w))
        n_samples, n_features = x.shape

        gradient = np.zeros((1, n_features), dtype=w.dtype)
        hessian = np.zeros((n_features, n_features), dtype=w.dtype)

        inv_n_samples = 1.0 / n_samples
        risk_set = 0
        risk_set_x = np.zeros((1, n_features), dtype=w.dtype)
        risk_set_xx = np.zeros((n_features, n_features), dtype=w.dtype)
        k = 0
        # iterate time in descending order
        while k < n_samples:
            ti = time[k]
            n_events = 0
            numerator = 0
            risk_set2 = 0
            risk_set_x2 = np.zeros_like(risk_set_x)
            risk_set_xx2 = np.zeros_like(risk_set_xx)
            while k < n_samples and ti == time[k]:
                # preserve 2D shape of row vector
                xk = x[k: k + 1]

                # outer product
                xx = np.dot(xk.T, xk)

                if self.event[k]:
                    numerator += xk
                    risk_set2 += exp_xw[k]
                    risk_set_x2 += exp_xw[k] * xk
                    risk_set_xx2 += exp_xw[k] * xx
                    n_events += 1
                else:
                    risk_set += exp_xw[k]
                    risk_set_x += exp_xw[k] * xk
                    risk_set_xx += exp_xw[k] * xx
                k += 1

            if n_events > 0:
                if breslow:
                    risk_set += risk_set2
                    risk_set_x += risk_set_x2
                    risk_set_xx += risk_set_xx2

                    z = risk_set_x / risk_set
                    gradient -= (numerator - n_events * z) * inv_n_samples

                    a = risk_set_xx / risk_set
                    # outer product
                    b = np.dot(z.T, z)

                    hessian += n_events * (a - b) * inv_n_samples
                else:
                    numerator /= n_events
                    for _ in range(n_events):
                        risk_set += risk_set2 / n_events
                        risk_set_x += risk_set_x2 / n_events
                        risk_set_xx += risk_set_xx2 / n_events

                        z = risk_set_x / risk_set
                        gradient -= (numerator - z) * inv_n_samples

                        a = risk_set_xx / risk_set
                        # outer product
                        b = np.dot(z.T, z)

                        hessian += (a - b) * inv_n_samples

        if not self.no_alpha:
            gradient += self.alpha * inv_n_samples * w

            diag_idx = np.diag_indices(n_features)
            hessian[diag_idx] += self.alpha * inv_n_samples

        self.gradient = gradient.ravel()
        self.hessian = hessian


class VerboseReporter:
    def __init__(self, verbose):
        self.verbose = verbose

    def end_max_iter(self, i):
        if self.verbose > 0:
            print(
                f"iter {i + 1:>6d}: reached maximum number of iterations. Stopping.")

    def end_converged(self, i):
        if self.verbose > 0:
            print(f"iter {i + 1:>6d}: optimization converged")

    def update(self, i, delta, loss_new):
        if self.verbose > 2:
            print(f"iter {i + 1:>6d}: update = {delta}")
        if self.verbose > 1:
            print(f"iter {i + 1:>6d}: loss = {loss_new:.10f}")

    def step_halving(self, i, loss):
        if self.verbose > 1:
            print(
                f"iter {i:>6d}: loss increased, performing step-halving. loss = {loss:.10f}")


class CoxPHSurvivalAnalysis(BaseEstimator, SurvivalAnalysisMixin):
    """Cox proportional hazards model.

    There are two possible choices for handling tied event times.
    The default is Breslow's method, which considers each of the
    events at a given time as distinct. Efron's method is more
    accurate if there are a large number of ties. When the number
    of ties is small, the estimated coefficients by Breslow's and
    Efron's method are quite close. Uses Newton-Raphson optimization.

    See [1]_, [2]_, [3]_ for further description.

    Parameters
    ----------
    alpha : float, ndarray of shape (n_features,), optional, default: 0
        Regularization parameter for ridge regression penalty.
        If a single float, the same penalty is used for all features.
        If an array, there must be one penalty for each feature.
        If you want to include a subset of features without penalization,
        set the corresponding entries to 0.

    ties : {'breslow', 'efron'}, optional, default: 'breslow'
        The method to handle tied event times. If there are
        no tied event times all the methods are equivalent.

    n_iter : int, optional, default: 100
        Maximum number of iterations.

    tol : float, optional, default: 1e-9
        Convergence criteria. Convergence is based on the negative log-likelihood::

        |1 - (new neg. log-likelihood / old neg. log-likelihood) | < tol

    verbose : int, optional, default: 0
        Specifies the amount of additional debug information
        during optimization.

    Attributes
    ----------
    coef_ : ndarray, shape = (n_features,)
        Coefficients of the model

    cum_baseline_hazard_ : :class:`sksurv.functions.StepFunction`
        Estimated baseline cumulative hazard function.

    baseline_survival_ : :class:`sksurv.functions.StepFunction`
        Estimated baseline survival function.

    n_features_in_ : int
        Number of features seen during ``fit``.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during ``fit``. Defined only when `X`
        has feature names that are all strings.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    See also
    --------
    sksurv.linear_model.CoxnetSurvivalAnalysis
        Cox proportional hazards model with l1 (LASSO) and l2 (ridge) penalty.

    References
    ----------
    .. [1] Cox, D. R. Regression models and life tables (with discussion).
           Journal of the Royal Statistical Society. Series B, 34, 187-220, 1972.
    .. [2] Breslow, N. E. Covariance Analysis of Censored Survival Data.
           Biometrics 30 (1974): 89–99.
    .. [3] Efron, B. The Efficiency of Cox’s Likelihood Function for Censored Data.
           Journal of the American Statistical Association 72 (1977): 557–565.
    """

    _parameter_constraints: dict = {
        "alpha": [Interval(numbers.Real, 0, None, closed="left"), np.ndarray],
        "ties": [StrOptions({"breslow", "efron"})],
        "n_iter": [Interval(numbers.Integral, 1, None, closed="left")],
        "tol": [Interval(numbers.Real, 0, None, closed="left")],
        "verbose": ["verbose"],
    }

    def __init__(self, alpha=0, *, ties="breslow", n_iter=100, tol=1e-9, verbose=0):
        self.alpha = alpha
        self.ties = ties
        self.n_iter = n_iter
        self.tol = tol
        self.verbose = verbose

        self._baseline_model = BreslowEstimator()

    @property
    def cum_baseline_hazard_(self):
        return self._baseline_model.cum_baseline_hazard_

    @property
    def baseline_survival_(self):
        return self._baseline_model.baseline_survival_

    @property
    def unique_times_(self):
        return self._baseline_model.unique_times_

    def fit(self, X, y):
        """Minimize negative partial log-likelihood for provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix

        y : structured array, shape = (n_samples,)
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        Returns
        -------
        self
        """
        self._validate_params()

        X = self._validate_data(X, ensure_min_samples=2, dtype=np.float64)
        event, time = check_array_survival(X, y)

        if isinstance(self.alpha, (numbers.Real, numbers.Integral)):
            alphas = np.empty(X.shape[1], dtype=float)
            alphas[:] = self.alpha
        else:
            alphas = self.alpha

        alphas = check_array(
            alphas,
            ensure_2d=False,
            ensure_min_samples=0,
            estimator=self,
            input_name="alpha",
        )
        if np.any(alphas < 0):
            raise ValueError(f"alpha must be positive, but was {self.alpha!r}")
        if alphas.shape[0] != X.shape[1]:
            raise ValueError(
                f"Length alphas ({alphas.shape[0]}) must match number of features ({X.shape[1]}).")

        optimizer = CoxPHOptimizer(X, event, time, alphas, self.ties)

        verbose_reporter = VerboseReporter(self.verbose)
        w = np.zeros(X.shape[1])
        w_prev = w
        i = 0
        loss = float("inf")
        while True:
            if i >= self.n_iter:
                verbose_reporter.end_max_iter(i)
                warnings.warn(
                    ("Optimization did not converge: Maximum number of iterations has been exceeded."),
                    stacklevel=2,
                    category=ConvergenceWarning,
                )
                break

            optimizer.update(w)
            delta = solve(
                optimizer.hessian,
                optimizer.gradient,
                overwrite_a=False,
                overwrite_b=False,
                check_finite=False,
            )

            if not np.all(np.isfinite(delta)):
                raise ValueError(
                    "search direction contains NaN or infinite values")

            w_new = w - delta
            loss_new = optimizer.nlog_likelihood(w_new)
            verbose_reporter.update(i, delta, loss_new)
            if loss_new > loss:
                # perform step-halving if negative log-likelihood does not decrease
                w = (w_prev + w) / 2
                loss = optimizer.nlog_likelihood(w)
                verbose_reporter.step_halving(i, loss)
                i += 1
                continue

            w_prev = w
            w = w_new

            res = np.abs(1 - (loss_new / loss))
            if res < self.tol:
                verbose_reporter.end_converged(i)
                break

            loss = loss_new
            i += 1

        self.coef_ = w
        self._baseline_model.fit(np.dot(X, self.coef_), event, time)
        return self

    def predict(self, X):
        """Predict risk scores.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix.

        Returns
        -------
        risk_score : array, shape = (n_samples,)
            Predicted risk scores.
        """
        check_is_fitted(self, "coef_")

        X = self._validate_data(X, reset=False)

        return np.dot(X, self.coef_)

    def predict_cumulative_hazard_function(self, X, return_array=False):
        """Predict cumulative hazard function.

        The cumulative hazard function for an individual
        with feature vector :math:`x` is defined as

        .. math::

            H(t \\mid x) = \\exp(x^\\top \\beta) H_0(t) ,

        where :math:`H_0(t)` is the baseline hazard function,
        estimated by Breslow's estimator.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix.

        return_array : boolean
            If set, return an array with the cumulative hazard rate
            for each `self.unique_times_`, otherwise an array of
            :class:`sksurv.functions.StepFunction`.

        Returns
        -------
        cum_hazard : ndarray
            If `return_array` is set, an array with the cumulative hazard rate
            for each `self.unique_times_`, otherwise an array of length `n_samples`
            of :class:`sksurv.functions.StepFunction` instances will be returned.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from sksurv.datasets import load_whas500
        >>> from sksurv.linear_model import CoxPHSurvivalAnalysis

        Load the data.

        >>> X, y = load_whas500()
        >>> X = X.astype(float)

        Fit the model.

        >>> estimator = CoxPHSurvivalAnalysis().fit(X, y)

        Estimate the cumulative hazard function for the first 10 samples.

        >>> chf_funcs = estimator.predict_cumulative_hazard_function(X.iloc[:10])

        Plot the estimated cumulative hazard functions.

        >>> for fn in chf_funcs:
        ...     plt.step(fn.x, fn(fn.x), where="post")
        ...
        >>> plt.ylim(0, 1)
        >>> plt.show()
        """
        return self._predict_cumulative_hazard_function(self._baseline_model, self.predict(X), return_array)

    def predict_survival_function(self, X, return_array=False):
        """Predict survival function.

        The survival function for an individual
        with feature vector :math:`x` is defined as

        .. math::

            S(t \\mid x) = S_0(t)^{\\exp(x^\\top \\beta)} ,

        where :math:`S_0(t)` is the baseline survival function,
        estimated by Breslow's estimator.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix.

        return_array : boolean, default: False
            If set, return an array with the probability
            of survival for each `self.unique_times_`,
            otherwise an array of :class:`sksurv.functions.StepFunction`.

        Returns
        -------
        survival : ndarray
            If `return_array` is set, an array with the probability of
            survival for each `self.unique_times_`, otherwise an array of
            length `n_samples` of :class:`sksurv.functions.StepFunction`
            instances will be returned.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from sksurv.datasets import load_whas500
        >>> from sksurv.linear_model import CoxPHSurvivalAnalysis

        Load the data.

        >>> X, y = load_whas500()
        >>> X = X.astype(float)

        Fit the model.

        >>> estimator = CoxPHSurvivalAnalysis().fit(X, y)

        Estimate the survival function for the first 10 samples.

        >>> surv_funcs = estimator.predict_survival_function(X.iloc[:10])

        Plot the estimated survival functions.

        >>> for fn in surv_funcs:
        ...     plt.step(fn.x, fn(fn.x), where="post")
        ...
        >>> plt.ylim(0, 1)
        >>> plt.show()
        """
        return self._predict_survival_function(self._baseline_model, self.predict(X), return_array)



import numpy as np
from numpy.testing import assert_almost_equal
import pandas as pd
from pandas.testing import assert_frame_equal
import pytest

from survivalist.compare import compare_survival
from survivalist.datasets import load_veterans_lung_cancer
from survivalist.testing import FixtureParameterFactory
from survivalist.util import Surv


def assert_stats_frame_equal(computed_stats, expected_stats):
    expected_stats["statistic"] = expected_stats["observed"] - expected_stats["expected"]
    assert_frame_equal(computed_stats, expected_stats)


def test_logrank_rossi(rossi):
    chisq, pval, stats, covar = compare_survival(rossi.y, rossi.x.loc[:, "race"], return_stats=True)

    expected_stats = pd.DataFrame(
        {
            "counts": np.array([53, 379], dtype=np.intp),
            "observed": np.array([12, 102]),
            "expected": np.array([14.709347908949, 99.2906520910509]),
        },
        index=pd.Index([0, 1], name="group", dtype=object),
    )
    assert_stats_frame_equal(stats, expected_stats)

    expected_var = np.array([[12.7417883289863, -12.7417883289863], [-12.7417883289863, 12.7417883289863]])
    assert_almost_equal(covar, expected_var)

    expected_chisq = 0.576101713683918
    expected_pval = 0.447844394252168

    assert chisq == pytest.approx(expected_chisq)
    assert pval == pytest.approx(expected_pval)


def test_logrank_veterans():
    X, y = load_veterans_lung_cancer()

    chisq, pval, stats, covar = compare_survival(y, X.loc[:, "Celltype"], return_stats=True)

    expected_stats = pd.DataFrame(
        columns=["counts", "observed", "expected", "statistic"],
        index=pd.Index(["adeno", "large", "smallcell", "squamous"], name="group"),
    )
    expected_stats["counts"] = np.array([27, 27, 48, 35], dtype=np.intp)
    expected_stats["observed"] = np.array([26, 26, 45, 31], dtype=int)
    expected_stats["expected"] = [15.6937646143605, 34.5494783863493, 30.1020793268148, 47.6546776724754]
    assert_stats_frame_equal(stats, expected_stats)

    expected_var = np.array(
        [
            [12.9661700605014, -4.07011754397142, -4.40872930298506, -4.48732321354496],
            [-4.07011754397142, 24.1990352938484, -7.81168661717217, -12.3172311327048],
            [-4.40872930298506, -7.81168661717217, 21.7542679406138, -9.53385202045655],
            [-4.48732321354496, -12.3172311327048, -9.53385202045655, 26.3384063667063],
        ]
    )
    assert_almost_equal(covar, expected_var)

    expected_chisq = 25.4037003457854
    expected_pval = 1.27124593900609e-05

    assert chisq == pytest.approx(expected_chisq)
    assert pval == pytest.approx(expected_pval)


class LogRankToyCases(FixtureParameterFactory):
    @property
    def time_1(self):
        return [12, 1, 6, 9, 7, 4, 13, 5, 2, 3]

    @property
    def time_2(self):
        return [12, 1, 2, 2, 7, 4, 1, 5, 2, 5]

    @property
    def two_groups(self):
        return np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])

    def data_two_groups_1(self):
        status = [False, True, True, False, False, False, True, False, True, True]
        y = Surv.from_arrays(status, self.time_1)

        expected_chisq = 0.0658861965695617
        expected_pval = 0.797423444741849
        return (y, self.two_groups), (expected_chisq, expected_pval)

    def data_two_groups_2(self):
        status = [False, False, True, False, False, False, True, False, True, True]
        y = Surv.from_arrays(status, self.time_1)

        expected_chisq = 0.843638413831986
        expected_pval = 0.358358251735077
        return (y, self.two_groups), (expected_chisq, expected_pval)

    def data_two_groups_3(self):
        status = [False, False, True, False, False, False, True, False, True, True]
        y = Surv.from_arrays(status, self.time_2)

        expected_chisq = 1.07692307692308
        expected_pval = 0.299386905803601
        return (y, self.two_groups), (expected_chisq, expected_pval)

    def data_two_groups_4(self):
        status = [False, False, True, True, False, False, False, True, False, True]
        y = Surv.from_arrays(status, self.time_2)

        expected_chisq = 0.0
        expected_pval = 1.0
        return (y, self.two_groups), (expected_chisq, expected_pval)

    def data_two_groups_5(self):
        status = [True, True, False, False, False, False, False, True, False, True]
        time = [12, 12, 2, 2, 7, 4, 1, 5, 2, 5]
        y = Surv.from_arrays(status, time)

        expected_chisq = 4.0
        expected_pval = 0.0455002638963585
        return (y, self.two_groups), (expected_chisq, expected_pval)

    def data_three_groups(self):
        time = [
            88,
            62,
            81,
            64,
            19,
            35,
            49,
            82,
            57,
            6,
            97,
            52,
            4,
            90,
            16,
            92,
            4,
            59,
            90,
            71,
            39,
            94,
            68,
            77,
            67,
            15,
            67,
            7,
            32,
            14,
            39,
            26,
            24,
            2,
            4,
            95,
            2,
            12,
            94,
            64,
            68,
            16,
            99,
            6,
            57,
            81,
            53,
            33,
            31,
            91,
            91,
        ]
        status = [
            False,
            False,
            True,
            False,
            False,
            False,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            True,
            False,
            True,
            True,
            False,
            False,
            False,
            True,
            False,
            False,
            True,
            False,
            False,
            True,
            False,
            True,
            False,
            True,
            False,
            False,
            True,
            False,
            False,
        ]
        y = Surv.from_arrays(status, time)
        x = np.array(
            [
                1,
                13,
                1,
                1,
                13,
                4,
                1,
                4,
                4,
                4,
                1,
                4,
                13,
                4,
                4,
                1,
                13,
                4,
                4,
                4,
                4,
                13,
                4,
                13,
                1,
                4,
                4,
                13,
                13,
                4,
                13,
                13,
                1,
                4,
                4,
                4,
                13,
                4,
                13,
                4,
                13,
                13,
                13,
                1,
                13,
                13,
                1,
                13,
                13,
                13,
                13,
            ]
        )

        expected_chisq = 0.0398155305699465
        expected_pval = 0.980289085821579
        return (y, x), (expected_chisq, expected_pval)

    def data_four_groups(self):
        time = [
            95,
            54,
            94,
            72,
            52,
            68,
            53,
            53,
            21,
            69,
            58,
            67,
            65,
            22,
            95,
            78,
            86,
            18,
            35,
            43,
            72,
            88,
            24,
            53,
            3,
            41,
            36,
            3,
            55,
            89,
            53,
            8,
            13,
            94,
            12,
            64,
            9,
            56,
            8,
            7,
        ]
        status = [
            False,
            False,
            True,
            False,
            False,
            True,
            False,
            True,
            False,
            True,
            False,
            True,
            False,
            True,
            True,
            True,
            True,
            True,
            False,
            True,
            False,
            True,
            False,
            True,
            True,
            False,
            False,
            True,
            True,
            False,
            True,
            False,
            True,
            True,
            False,
            False,
            True,
            True,
            True,
            False,
        ]
        y = Surv.from_arrays(status, time)
        x = np.array(
            [
                3,
                3,
                2,
                -1,
                3,
                -1,
                2,
                0,
                -1,
                2,
                2,
                -1,
                -1,
                0,
                2,
                2,
                -1,
                -1,
                -1,
                3,
                -1,
                -1,
                2,
                -1,
                3,
                2,
                -1,
                2,
                0,
                2,
                2,
                2,
                2,
                3,
                -1,
                2,
                -1,
                2,
                2,
                2,
            ]
        )

        expected_chisq = 5.25844160740909
        expected_pval = 0.153821897350625
        return (y, x), (expected_chisq, expected_pval)


@pytest.mark.parametrize("inputs,expectation", LogRankToyCases().get_cases())
def test_toy_compare_survival(inputs, expectation):
    chisq, pval = compare_survival(*inputs)

    expected_chisq, expected_pval = expectation
    assert chisq == pytest.approx(expected_chisq)
    assert pval == pytest.approx(expected_pval)


class LogRankFailureCases(FixtureParameterFactory):
    def data_single_group(self):
        y = Surv.from_arrays([True] * 7, np.arange(1, 8))
        group = np.ones(7)

        err = pytest.raises(ValueError, match=r"At least two groups must be specified, but only one was provided\.")
        return (y, group), err

    def data_wrong_shape(self):
        y = Surv.from_arrays([True] * 7, np.arange(1, 8))
        group = np.empty((y.shape[0], 2, 4), dtype="str")
        group[:, 0, :] = "a"
        group[:, 1, :] = "b"

        err = pytest.raises(ValueError, match=r"Found array with dim 3\. compare_survival expected <= 2\.")
        return (y, group), err


@pytest.mark.parametrize("inputs,excepted_error", LogRankFailureCases().get_cases())
def test_compare_survival_failures(inputs, excepted_error):
    with excepted_error:
        compare_survival(*inputs)



import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
import pytest
from sklearn.exceptions import ConvergenceWarning
from sklearn.preprocessing import scale

from survivalist.column import encode_categorical
from survivalist.datasets import load_gbsg2
from survivalist.exceptions import NoComparablePairException
from survivalist.svm._minlip import create_difference_matrix
from survivalist.svm.minlip import HingeLossSurvivalSVM, MinlipSurvivalAnalysis
from survivalist.testing import FixtureParameterFactory, assert_cindex_almost_equal
from survivalist.util import Surv


def create_toy_data():
    x = np.array(
        [
            [1.0, 1.0],
            [10.2, 15.0],
            [20.0, 5.0],
            [40, 30],
            [45, 21],
            [50, 36],
        ]
    )

    t = np.random.RandomState(0).exponential(scale=8, size=x.shape[0])
    t.sort()
    y = Surv.from_arrays(
        [True, True, False, True, False, False],
        t,
        name_event="status",
    )
    return x, y


@pytest.fixture()
def toy_data():
    return create_toy_data()


@pytest.fixture()
def toy_test_data():
    x = np.array(
        [
            [1.0, 1.0],
            [40, 30],
            [50, 36],
        ]
    )
    rnd = np.random.RandomState(136)
    x += rnd.randn(*x.shape)
    return x


@pytest.fixture()
def gbsg2():
    x, y = load_gbsg2()
    x = encode_categorical(x)
    return x.values, y


class DifferenceMatrixToyDataCases(FixtureParameterFactory):
    @property
    def time_and_event(self):
        _, y = create_toy_data()
        time = y["time"]
        status = y["status"]
        return time, status

    def _create_expected(self, shape=(5, 6)):
        return np.zeros(shape, dtype=np.int8)

    def data_direct_neighbor_without_censoring(self):
        kind = "next"
        time, status = self.time_and_event
        status = np.ones_like(status)

        expected = self._create_expected()
        expected[0, 0] = -1
        expected[0, 1] = 1
        expected[1, 1] = -1
        expected[1, 2] = 1
        expected[2, 2] = -1
        expected[2, 3] = 1
        expected[3, 3] = -1
        expected[3, 4] = 1
        expected[4, 4] = -1
        expected[4, 5] = 1

        expected_diff = time[1:] - time[:-1]

        return time, status.astype(np.uint8), kind, expected, expected_diff

    def data_direct_neighbor_without_censoring_shuffled(self):
        kind = "next"
        time, status = self.time_and_event
        status = np.ones_like(status)
        order = [3, 2, 5, 0, 1, 4]  # = [ 20.  11.  70.   3.   6.  37.]

        expected = self._create_expected()
        expected[0, 3] = -1
        expected[0, 4] = 1
        expected[1, 4] = -1
        expected[1, 1] = 1
        expected[2, 1] = -1
        expected[2, 0] = 1
        expected[3, 0] = -1
        expected[3, 5] = 1
        expected[4, 5] = -1
        expected[4, 2] = 1

        expected_diff = time[1:] - time[:-1]
        time = time[order]

        return time, status.astype(np.uint8), kind, expected, expected_diff

    def data_direct_neighbor_with_censoring(self):
        kind = "next"
        time, status = self.time_and_event

        expected = self._create_expected((3, 6))
        expected[0, 0] = -1
        expected[0, 1] = 1
        expected[1, 1] = -1
        expected[1, 2] = 1
        expected[2, 3] = -1
        expected[2, 4] = 1

        comparable_pairs = np.array([True, True, False, True, False])
        expected_diff = (time[1:] - time[:-1])[comparable_pairs]

        return time, status.astype(np.uint8), kind, expected, expected_diff

    def data_nearest_neighbor(self):
        kind = "nearest"
        time, status = self.time_and_event
        status = np.ones_like(status)

        expected = self._create_expected()
        expected[0, 0] = -1
        expected[0, 1] = 1
        expected[1, 1] = -1
        expected[1, 2] = 1
        expected[2, 2] = -1
        expected[2, 3] = 1
        expected[3, 3] = -1
        expected[3, 4] = 1
        expected[4, 4] = -1
        expected[4, 5] = 1

        return time, status.astype(np.uint8), kind, expected, None

    def data_nearest_neighbor_censored(self):
        kind = "nearest"
        time, status = self.time_and_event

        expected = self._create_expected()
        expected[0, 0] = -1
        expected[0, 1] = 1
        expected[1, 1] = -1
        expected[1, 2] = 1
        expected[2, 1] = -1
        expected[2, 3] = 1
        expected[3, 3] = -1
        expected[3, 4] = 1
        expected[4, 3] = -1
        expected[4, 5] = 1

        return time, status.astype(np.uint8), kind, expected, None

    def data_full(self):
        kind = "all"
        time, status = self.time_and_event
        status = np.ones_like(status)

        expected = self._create_expected((15, 6))
        expected[0, 1] = 1
        expected[0, 0] = -1

        expected[1:3, 2] = 1
        expected[1, 0] = -1
        expected[2, 1] = -1

        expected[3:6, 3] = 1
        expected[3, 0] = -1
        expected[4, 1] = -1
        expected[5, 2] = -1

        expected[6:10, 4] = 1
        expected[6, 0] = -1
        expected[7, 1] = -1
        expected[8, 2] = -1
        expected[9, 3] = -1

        expected[10:15, 5] = 1
        expected[10, 0] = -1
        expected[11, 1] = -1
        expected[12, 2] = -1
        expected[13, 3] = -1
        expected[14, 4] = -1

        return time, status.astype(np.uint8), kind, expected, None

    def data_full_censored(self):
        kind = "all"
        time, status = self.time_and_event

        expected = self._create_expected((11, 6))
        expected[0, 1] = 1
        expected[0, 0] = -1

        expected[1:3, 2] = 1
        expected[1, 0] = -1
        expected[2, 1] = -1

        expected[3:5, 3] = 1
        expected[3, 0] = -1
        expected[4, 1] = -1

        expected[5:8, 4] = 1
        expected[5, 0] = -1
        expected[6, 1] = -1
        expected[7, 3] = -1

        expected[8:12, 5] = 1
        expected[8, 0] = -1
        expected[9, 1] = -1
        expected[10, 3] = -1

        return time, status.astype(np.uint8), kind, expected, None


@pytest.mark.parametrize("time,status,kind,expected_mat,expected_diff", DifferenceMatrixToyDataCases().get_cases())
def test_toy_create_difference_matrix(time, status, kind, expected_mat, expected_diff):
    mat = create_difference_matrix(status, time, kind=kind)

    assert_array_equal(expected_mat, mat.toarray())

    if expected_diff is not None:
        # should return first-order differences of y["time"]
        actual_diff = mat.dot(time)
        assert_array_almost_equal(expected_diff, actual_diff)


@pytest.fixture()
def minlip_model_factory():
    def create_and_fit_model(solver, x, y, **kwargs):
        params = {"solver": solver, "alpha": 1, "pairs": "next"}
        params.update(kwargs)
        return MinlipSurvivalAnalysis(**params).fit(x, y)

    return create_and_fit_model


@pytest.fixture()
def toy_data_standardized(toy_data, toy_test_data):
    x, y = toy_data
    m = np.mean(x, axis=0)
    sd = np.std(x, axis=0)
    return ((x - m) / sd, y), (toy_test_data - m) / sd


class TestToyMinlipSurvivalAnalysis:
    @staticmethod
    @pytest.mark.parametrize("solver,expected_iters", [("osqp", 100), ("ecos", 10)])
    def test_fit_alpha_2(minlip_model_factory, toy_data, solver, expected_iters):
        x, y = toy_data
        m = minlip_model_factory(solver, x, y, alpha=2.0)

        assert m.n_iter_ > expected_iters
        assert (1, 6) == m.coef_.shape
        assert 1 == m.coef0
        expected_coef = np.array(
            [[-0.011728003147, 0.011728002895, 0.000000000252, -0.017524801335, 0.017524801335, 0.0]]
        )
        assert_array_almost_equal(m.coef_, expected_coef)

    @staticmethod
    @pytest.mark.parametrize("solver", ["osqp", "ecos"])
    def test_timeit(minlip_model_factory, toy_data, solver):
        x, y = toy_data
        m = minlip_model_factory(solver, x, y, timeit=7)

        assert 7 == len(m.timings_)

    @staticmethod
    @pytest.mark.parametrize("solver", ["osqp", "ecos"])
    def test_predict_1(minlip_model_factory, toy_data, solver):
        x, y = toy_data
        p = minlip_model_factory(solver, x, y).predict(x)
        assert_cindex_almost_equal(y["status"], y["time"], p, (1.0, 11, 0, 0, 0))

    @staticmethod
    @pytest.mark.parametrize("solver", ["osqp", "ecos"])
    def test_predict_2(minlip_model_factory, toy_data_standardized, solver):
        (x, y), x_test = toy_data_standardized

        y = y.copy()
        y["time"] = np.arange(1, 7)

        p = minlip_model_factory(solver, x, y, pairs="next").predict(x_test)

        expected = np.array([1.368221203557392, -0.476483331099142, -1.009079072163642])
        assert_array_almost_equal(expected, p, decimal=5)


@pytest.fixture()
def svm_model_factory():
    def create_and_fit_model(solver, x, y, **kwargs):
        params = {"solver": solver, "alpha": 2}
        params.update(kwargs)
        return HingeLossSurvivalSVM(**params).fit(x, y)

    return create_and_fit_model


@pytest.mark.parametrize("solver", ["osqp", "ecos"])
@pytest.mark.parametrize("pairs", ["all", "nearest"])
def test_toy_hinge_fit_and_predict(svm_model_factory, toy_data_standardized, solver, pairs):
    (x, y), x_test = toy_data_standardized
    m = svm_model_factory(solver, x, y, pairs=pairs)

    assert (1, 6) == m.coef_.shape
    assert 1 == m.coef0
    expected_coef = np.array([[-1.893832101337, 1.083653895940, 0.810178205398, -2.0, 2.0, 0.0]])
    assert_array_almost_equal(m.coef_, expected_coef, decimal=5)

    p = m.predict(x)
    assert_cindex_almost_equal(y["status"], y["time"], p, (1.0, 11, 0, 0, 0))

    p = m.predict(x_test)
    expected = np.array([2.8571060045, -1.2661069033, -2.3044907774])
    assert_array_almost_equal(expected, p, decimal=5)


@pytest.fixture()
def gbsg2_scaled(gbsg2):
    x, y = gbsg2
    x = scale(x)
    return x, y


class TestMinlipBreastCancer:
    @staticmethod
    @pytest.mark.parametrize("solver,expected_iters", [("osqp", 1000), ("ecos", 10)])
    def test_fit_and_predict(gbsg2_scaled, minlip_model_factory, solver, expected_iters):
        x, y = gbsg2_scaled
        m = minlip_model_factory(solver, x, y)

        assert (1, x.shape[0]) == m.coef_.shape

        assert m.n_iter_ > expected_iters

        p = m.predict(x)
        assert_cindex_almost_equal(y["cens"], y["time"], p, (0.5990741854033906, 79720, 53352, 0, 42))

    @staticmethod
    @pytest.mark.slow()
    @pytest.mark.parametrize(
        "solver,expected_cindex",
        [
            ("osqp", (0.6106092942166647, 81255, 51817, 0, 42)),
            ("ecos", (0.6105867500300589, 81252, 51820, 0, 42)),
        ],
    )
    def test_fit_and_predict_rbf(gbsg2_scaled, minlip_model_factory, solver, expected_cindex):
        x, y = gbsg2_scaled
        m = minlip_model_factory(solver, x, y, kernel="rbf", gamma=1.0 / 8, pairs="next", max_iter=1000)
        m.fit(x, y)

        assert (1, x.shape[0]) == m.coef_.shape

        p = m.predict(x)
        assert_cindex_almost_equal(y["cens"], y["time"], p, expected_cindex)

    @staticmethod
    @pytest.mark.slow()
    @pytest.mark.parametrize("solver", ["osqp", "ecos"])
    def test_kernel_precomputed(gbsg2_scaled, solver):
        x, y = gbsg2_scaled
        from sklearn.metrics.pairwise import pairwise_kernels
        from sklearn.utils.metaestimators import _safe_split

        m = MinlipSurvivalAnalysis(solver=solver, kernel="precomputed", max_iter=25000)
        K = pairwise_kernels(x, metric="rbf", gamma=0.1)

        train_idx = np.arange(200, x.shape[0])
        test_idx = np.arange(200)
        X_fit, y_fit = _safe_split(m, K, y, train_idx)
        X_test, y_test = _safe_split(m, K, y, test_idx, train_idx)

        m.fit(X_fit, y_fit)

        p = m.predict(X_test)
        assert_cindex_almost_equal(y_test["cens"], y_test["time"], p, (0.6518928901200369, 8472, 4524, 0, 3))

    @staticmethod
    @pytest.mark.parametrize("solver", ["osqp", "ecos"])
    def test_max_iter(gbsg2_scaled, solver):
        x, y = gbsg2_scaled
        m = MinlipSurvivalAnalysis(solver=solver, alpha=1, kernel="poly", degree=2, pairs="next", max_iter=5)

        with pytest.warns(
            ConvergenceWarning, match=f"{solver.upper()} solver did not converge: maximum iterations reached"
        ):
            m.fit(x, y)


@pytest.mark.parametrize(
    "solver",
    [
        None,
        "i don't know",
        [("why", "are"), ("you", "doing this")],
    ],
)
def test_unknown_solver(gbsg2, solver):
    x, y = gbsg2
    m = MinlipSurvivalAnalysis(solver=solver)

    msg = r"The 'solver' parameter of MinlipSurvivalAnalysis must be a str among \{.+\}\. Got .+ instead\."
    with pytest.raises(ValueError, match=msg):
        m.fit(x, y)


@pytest.mark.parametrize("model_cls", [MinlipSurvivalAnalysis, HingeLossSurvivalSVM])
@pytest.mark.parametrize("solver", ["ecos", "osqp"])
@pytest.mark.parametrize("pairs", ["all", "nearest", "next"])
def test_fit_uncomparable(whas500_uncomparable, model_cls, solver, pairs):
    ssvm = model_cls(solver=solver, pairs=pairs)
    with pytest.raises(NoComparablePairException):
        ssvm.fit(whas500_uncomparable.x, whas500_uncomparable.y)



import numpy as np

from collections import namedtuple
from sklearn.calibration import CalibratedClassifierCV
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from .transformer import SurvivalStacker
from ..util import check_array_survival
from ..utils import simulate_replications
from ..base import SurvivalAnalysisMixin
from ..linear_model.coxph import BreslowEstimator
from ..util import check_array_survival
from ..ensemble.survival_loss import (
    LOSS_FUNCTIONS,
    CensoredSquaredLoss,
    CoxPH,
    IPCWLeastSquaresError,
)
from ..functions import StepFunction


class SurvStacker(SurvivalAnalysisMixin):
    """
    A class to create a Survival Stacker for any classifier.
    """

    def __init__(
        self, clf=RandomForestClassifier(), loss="squared", type_sim="none", replications=250, random_state=42, **kwargs
    ):
        """
        Parameters
        ----------
        clf : classifier, default: LogisticRegression()
            The classifier to be used for stacking.

        loss : {'coxph', 'squared', 'ipcwls'}, optional, default: 'coxph'
        loss function to be optimized. 'coxph' refers to partial likelihood loss
        of Cox's proportional hazards model. The loss 'squared' minimizes a
        squared regression loss that ignores predictions beyond the time of censoring,
        and 'ipcwls' refers to inverse-probability of censoring weighted least squares error.

        type_sim (str): Method for simulation:
                      - 'none': No simulation.
                      - 'kde': Kernel Density Estimation.
                      - 'bootstrap': Bootstrap resampling.
                      - 'normal': Parametric distribution fitting.
                      - 'ecdf': Empirical CDF-based sampling.
                      - 'permutation': Permutation resampling.
                      - 'smooth_bootstrap': Smoothed bootstrap with added noise.

        replications : int, default: 250
            The number of replications for the simulation.

        random_state : int seed, RandomState instance, or None, default: None
            The seed of the pseudo random number generator to use when
            shuffling the data.

        kwargs : additional parameters to be passed to CalibratedClassifierCV
        """
        self.random_state = random_state
        self.clf = clf
        try:
            self.clf.set_params(random_state=self.random_state)
        except Exception as e:
            pass
        self.clf = CalibratedClassifierCV(clf, cv=3, **kwargs)
        self.ss = SurvivalStacker()
        self._baseline_model = None
        self.loss = loss
        self._loss = LOSS_FUNCTIONS[self.loss]()
        self.type_sim = type_sim
        if self.type_sim not in ["none", "kde", "bootstrap", "normal", "ecdf", "permutation", "smooth_bootstrap"]:
            raise ValueError(
                f"Invalid type_sim value: {self.type_sim}. "
                "Choose from 'none', 'kde', 'bootstrap', 'normal', 'ecdf', 'permutation', or 'smooth_bootstrap'."
            )
        if self.loss not in ["coxph", "squared", "ipcwls"]:
            raise ValueError(
                f"Invalid loss value: {self.loss}. " "Choose from 'coxph', 'squared', or 'ipcwls'.")
        self.replications = replications
        self.times_ = None
        self.unique_times_ = None
        self.calibrated_residuals_ = None

    def _get_baseline_model(self):
        """
        Get the baseline model for the survival stacker.

        Returns
        -------
        self.ss : SurvivalStacker
            The fitted survival stacker.
        """
        return self._baseline_model

    def _set_baseline_model(self, X, event, time):
        if isinstance(self._loss, CoxPH):
            risk_scores = self.predict(X)
            self._baseline_model = BreslowEstimator().fit(risk_scores, event, time)
        else:
            self._baseline_model = None

    def fit(self, X, y, **kwargs):
        """
        Fit the Survival Stacker to the data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        y : array-like, shape (n_samples,)
            The target values (survival times).

        kwargs : additional parameters to be passed to the fitting function of the classifier (e.g., `sample_weight`)

        Returns
        -------
        self : object
            Returns self.
        """
        # Convert X to numpy array if needed
        if hasattr(X, "to_numpy"):
            X = X.to_numpy()

        # print("X shape:", X.shape)

        # Get survival stacker predictions
        X_oo, y_oo = self.ss.fit_transform(X, y)
        self.times_ = self.ss.times
        self.unique_times_ = np.sort(np.unique(self.ss.times))
        # print("X_oo shape:", X_oo.shape)
        # print("y_oo shape:", y_oo.shape)
        # print("self.times_ shape:", self.times_.shape)
        # print("self.unique_times_ shape:", self.unique_times_.shape)

        if self.type_sim != "none":
            half_n = X_oo.shape[0] // 2
            X_train_oo, X_calib_oo, y_train_oo, y_calib_oo = train_test_split(
                X_oo, y_oo, test_size=half_n, random_state=self.random_state, stratify=y_oo
            )
            # Fit classifier
            self.clf.fit(X_train_oo, y_train_oo, **kwargs)
            # print("y_train_oo:", y_train_oo)
            # Calibrate classifier
            calib_probs = self.clf.predict_proba(X_calib_oo)[:, 1]
            # print("calib_probs shape:", calib_probs.shape)
            encoder = OneHotEncoder(sparse_output=False)
            # print("calib response shape:", y_calib_oo.shape)
            test_probs = encoder.fit_transform(y_calib_oo.reshape(-1, 1))
            # print("test_probs shape:", test_probs.shape)
            self.calibrated_residuals_ = test_probs[:, 1] - calib_probs
            self.clf.fit(X_calib_oo, y_calib_oo, **kwargs)
            # Set baseline model
            event, time = check_array_survival(X, y)
            self._set_baseline_model(X, event, time)
            return self

        # Set baseline model
        self.clf.fit(X_oo, y_oo, **kwargs)
        event, time = check_array_survival(X, y)
        self._set_baseline_model(X, event, time)
        return self

    def _predict_survival_function_temp(self, X):
        """
        Predict survival function with normalized probabilities.
        """
        # print("X shape:", X.shape)
        X_risk, _ = self.ss.transform(X)
        # print("X_risk shape:", X_risk.shape)
        oo_test_estimates = self.clf.predict_proba(X_risk)[:, 1]
        # print("oo_test_estimates shape:", oo_test_estimates.shape)

        if self.type_sim == "none":
            # If no simulation, return the test estimates
            return self.ss.predict_survival_function(oo_test_estimates)

        # Apply the calibrated residuals
        # print("Calibrated residuals shape:", self.calibrated_residuals_.shape)
        # print("oo_test_estimates shape:", oo_test_estimates.shape)
        # Simulate the residuals
        simulations_oo_test_estimates = simulate_replications(
            data=self.calibrated_residuals_,
            num_replications=self.replications,
            n_obs=oo_test_estimates.shape[0],
            method=self.type_sim,
        )
        # print("simulations_oo_test_estimates shape:", simulations_oo_test_estimates.shape)
        # Add the calibrated residuals to the test estimates
        oo_test_estimates = np.tile(
            oo_test_estimates, (self.replications, 1)).T + simulations_oo_test_estimates
        # clip values to be between 0 and 1
        oo_test_estimates = np.clip(oo_test_estimates, 0, 1)

        # Normalize probabilities to sum to 1 for each replication
        for i in range(self.replications):
            probs = oo_test_estimates[:, i]
            probs = np.maximum(probs, 0)  # Ensure probabilities are positive
            probs = probs / np.sum(probs)  # Normalize to sum to 1
            oo_test_estimates[:, i] = probs

        return [self.ss.predict_survival_function(oo_test_estimates[:, i]) for i in range(self.replications)]

    def predict(self, X, threshold=0.5):

        surv = self._predict_survival_function_temp(X)

        if self.type_sim == "none":
            crossings = surv <= threshold  # Boolean array of threshold crossings
            # For each sample, get the index of the first crossing (or -1 if none)
            cross_indices = np.argmax(crossings, axis=1)
            # Handle cases where survival never crosses the threshold:
            # argmax returns 0 if no True found, so we need to check if the crossing is valid
            valid_crossings = crossings[np.arange(
                len(crossings)), cross_indices]
            # Map to actual times
            predicted_times = np.where(
                valid_crossings,
                self.unique_times_[cross_indices],
                self.unique_times_[-1],  # use max time if no crossing
            )
            return predicted_times

        crossings = [s <= threshold for s in surv]
        cross_indices = [np.argmax(cross) for cross in crossings]
        # Handle cases where survival never crosses the threshold:
        # argmax returns 0 if no True found, so we need to check if the crossing is valid
        valid_crossings = [cross[ci]
                           for cross, ci in zip(crossings, cross_indices)]
        # Map to actual times
        predicted_times = [
            # use max time if no crossing
            np.where(valid, self.unique_times_[ci], self.unique_times_[-1])
            for valid, ci in zip(valid_crossings, cross_indices)
        ]
        return np.array(predicted_times)

    def predict_cumulative_hazard_function(self, X, return_array=False):
        """
        Predict cumulative hazard function.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.
        return_array : bool, default=False
            Whether to return the cumulative hazard function as an array.
        Returns
        -------
        array-like or list of StepFunction
            Predicted cumulative hazard function for each sample.
        """
        if self.type_sims == "none":
            # Convert X to numpy array if needed
            return self._predict_cumulative_hazard_function(self._get_baseline_model(), self.predict(X), return_array)
        return [
            self._predict_cumulative_hazard_function(
                self._get_baseline_model(), s, return_array)
            for s in self._predict_survival_function_temp(X)
        ]

    def predict_survival_function(self, X, return_array=False, level=95):
        """
        Predict survival function.
        """
        if hasattr(X, "to_numpy"):
            X = X.to_numpy()

        surv = self._predict_survival_function_temp(X)

        # print("surv: ", surv)

        if self.type_sim == "none":
            if return_array:
                return surv

            funcs = []
            surv = np.asarray(surv)
            if surv.ndim == 1:
                surv = surv.reshape(1, -1)

            for i in range(surv.shape[0]):
                if len(self.unique_times_) != len(surv[i]):
                    # Créer une grille temporelle appropriée
                    x_old = np.linspace(0, 1, len(surv[i]))
                    x_new = np.linspace(0, 1, len(self.unique_times_))
                    surv_interp = np.interp(x_new, x_old, surv[i])
                else:
                    surv_interp = surv[i]
                func = StepFunction(x=self.unique_times_, y=surv_interp)
                funcs.append(func)
            return np.array(funcs)
        else:
            SurvivalCurves = namedtuple(
                "SurvivalCurves", ["mean", "lower", "upper"])
            # `surv` contains `replications` number of survival functions,
            # for each sample in the test set
            n_obs = X.shape[0]
            results = {}
            for j in range(n_obs):
                key = "obs" + str(j)
                results[key] = []
                for i in range(self.replications):
                    results[key].append(surv[i][j])
            # Calculate mean, lower and upper bounds
            mean_surv = []
            lower_surv = []
            upper_surv = []
            for key in results.keys():
                mean_surv.append(np.mean(results[key], axis=0))
                lower_surv.append(np.percentile(
                    results[key], (100 - level) / 2, axis=0))
                upper_surv.append(np.percentile(
                    results[key], 100 - (100 - level) / 2, axis=0))
            mean_surv = np.array(mean_surv)
            lower_surv = np.array(lower_surv)
            upper_surv = np.array(upper_surv)
            if return_array:
                return mean_surv, lower_surv, upper_surv
            # Create StepFunction objects for mean, lower and upper bounds
            mean_funcs = []
            lower_funcs = []
            upper_funcs = []
            for i in range(mean_surv.shape[0]):
                mean_func = StepFunction(x=self.unique_times_, y=mean_surv[i])
                lower_func = StepFunction(
                    x=self.unique_times_, y=lower_surv[i])
                upper_func = StepFunction(
                    x=self.unique_times_, y=upper_surv[i])
                mean_funcs.append(mean_func)
                lower_funcs.append(lower_func)
                upper_funcs.append(upper_func)
            return SurvivalCurves(mean=mean_funcs, lower=lower_funcs, upper=upper_funcs)



import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd
import pandas.testing as tm
import pytest

from survivalist import column
from survivalist.testing import FixtureParameterFactory


class StandardizeCase(FixtureParameterFactory):
    @property
    def numeric_data(self):
        return pd.DataFrame(np.arange(50, dtype=float).reshape(10, 5))

    @property
    def expected(self):
        return pd.DataFrame(
            np.array(
                [
                    [-1.486301, -1.486301, -1.486301, -1.486301, -1.486301],
                    [-1.156012, -1.156012, -1.156012, -1.156012, -1.156012],
                    [-0.825723, -0.825723, -0.825723, -0.825723, -0.825723],
                    [-0.495434, -0.495434, -0.495434, -0.495434, -0.495434],
                    [-0.165145, -0.165145, -0.165145, -0.165145, -0.165145],
                    [0.165145, 0.165145, 0.165145, 0.165145, 0.165145],
                    [0.495434, 0.495434, 0.495434, 0.495434, 0.495434],
                    [0.825723, 0.825723, 0.825723, 0.825723, 0.825723],
                    [1.156012, 1.156012, 1.156012, 1.156012, 1.156012],
                    [1.486301, 1.486301, 1.486301, 1.486301, 1.486301],
                ]
            )
        )

    @property
    def non_numeric_data(self):
        data = pd.DataFrame.from_dict(
            {
                "q1": ["no", "no", "yes", "yes", "no", "no", None, "yes", "no", None],
                "q2": ["maybe", "no", "yes", "maybe", "yes", "no", None, "maybe", "no", "yes"],
                "q3": [1, 2, 1, 3, 1, 2, np.nan, np.nan, 3, 2],
            }
        )

        data["q3"] = data.loc[:, "q3"].astype("category")
        return data

    def data_numeric(self):
        return self.numeric_data, self.expected

    def data_float_numpy_array(self):
        return self.numeric_data.values, self.expected

    def data_int_numpy_array(self):
        return self.numeric_data.values.astype(int), self.expected

    def data_non_numeric(self):
        return self.non_numeric_data, self.non_numeric_data

    def data_non_numeric_numpy_array(self):
        data = self.non_numeric_data.values
        return data, pd.DataFrame(data)

    def data_mixed(self):
        mixed_data_frame = pd.concat((self.numeric_data, self.non_numeric_data), axis=1)
        expected = pd.concat((self.expected, self.non_numeric_data), axis=1)
        return mixed_data_frame, expected

    def data_mixed_numpy_array(self):
        data, _ = self.data_mixed()
        data = data.values
        return data, pd.DataFrame(data)


@pytest.mark.parametrize("in_data,expected", StandardizeCase().get_cases())
def test_standardize(in_data, expected):
    before = in_data.copy()
    result = column.standardize(in_data)

    # check that data wasn't modified inplace
    if isinstance(before, np.ndarray) and np.issubdtype(before.dtype, float):
        assert_array_equal(before, in_data)
    elif isinstance(before, pd.DataFrame):
        tm.assert_frame_equal(before, in_data)

    if isinstance(result, np.ndarray):
        result = pd.DataFrame(result, columns=expected.columns)

    tm.assert_frame_equal(pd.isnull(result), pd.isnull(expected))
    tm.assert_frame_equal(result, expected)


class CategoricalCases(FixtureParameterFactory):
    def _make_randn(self, shape):
        return np.random.RandomState(0).randn(shape)

    @property
    def mixed_data_frame(self):
        a = np.r_[np.repeat(["large"], 10), np.repeat(["small"], 5), np.repeat(["tiny"], 13), np.repeat(["medium"], 3)]
        b = np.r_[np.repeat(["yes"], 8), np.repeat(["no"], 23)]

        c = self._make_randn(len(a))

        df = pd.DataFrame.from_dict({"a_category": a, "a_binary": b, "a_number": c.copy()})
        return df


class EncodeCategoricalCases(CategoricalCases):
    @property
    def binary_with_missing(self):
        inputs = np.r_[
            np.repeat(["yes"], 5),
            np.repeat([None], 10),
            np.repeat(["no"], 16),
        ]
        expected = np.r_[
            np.repeat([1.0], 5),
            np.repeat([np.nan], 10),
            np.repeat([0.0], 16),
        ]
        return inputs, expected

    def data_series_categorical(self):
        input_series = pd.Series(
            pd.Categorical.from_codes(
                [1, 1, 0, 2, 0, 1, 2, 1, 2, 0, 0, 1, 2, 2], ["small", "medium", "large"], ordered=False
            ),
            name="a_series",
        )

        expected_df = pd.DataFrame.from_dict(
            {
                "a_series=medium": np.array([1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0], dtype=float),
                "a_series=large": np.array([0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1], dtype=float),
            }
        )

        return input_series, {}, expected_df

    def data_case_1(self):
        input_df = self.mixed_data_frame

        eb = np.r_[np.repeat([1.0], 8), np.repeat([0.0], 23)]

        a_tiny = np.zeros(31, dtype=float)
        a_tiny[15:28] = 1

        a_small = np.zeros(31, dtype=float)
        a_small[10:15] = 1

        a_medium = np.zeros(31, dtype=float)
        a_medium[-3:] = 1

        expected_df = pd.DataFrame.from_dict(
            {
                "a_category=medium": a_medium,
                "a_category=small": a_small,
                "a_category=tiny": a_tiny,
                "a_binary=yes": eb,
                "a_number": input_df.loc[:, "a_number"].values.copy(),
            }
        )

        return input_df, {}, expected_df

    def data_duplicate_index(self):
        input_df = self.mixed_data_frame.drop("a_binary", axis=1)
        input_df = pd.concat((input_df.iloc[:11], input_df.iloc[[11]], input_df.iloc[11:]))

        index = np.ceil(np.arange(0, input_df.shape[0] // 2, 0.5))
        input_df.index = index

        expected_df = pd.DataFrame(
            np.zeros((32, 3), dtype=float),
            index=index,
            columns=["a_category=medium", "a_category=small", "a_category=tiny"],
        )
        # tiny
        expected_df.iloc[16:29, 2] = 1
        # small
        expected_df.iloc[10:16, 1] = 1
        # medium
        expected_df.iloc[-3:, 0] = 1

        expected_df.loc[:, "a_number"] = input_df.loc[:, "a_number"].values.copy()

        return input_df, {}, expected_df

    def data_numeric(self):
        a = np.array([0, 1, 1, 0, 1, 0, 0, 1, 0, 1], dtype=object)
        b = np.array([1, 2, 1, 3, 2, 1, 3, 2, 3, 1], dtype=object)
        c = np.array(
            [1.0 / 128, 1.0 / 32, 1.0, 1.0 / 8, 1.0 / 32, 1.0, 1.0 / 128, 1.0 / 8, 1.0, 1.0 / 32], dtype=object
        )

        input_df = pd.DataFrame.from_dict({"a_binary_int": a.copy(), "a_three_int": b.copy(), "a_four_float": c.copy()})

        expected_df = pd.DataFrame(
            {
                "a_binary_int=1": a.astype(float),
                "a_three_int=2": (b == 2).astype(float),
                "a_three_int=3": (b == 3).astype(float),
                f"a_four_float={1.0 / 32}": (c == 1.0 / 32).astype(float),
                f"a_four_float={1.0 / 8}": (c == 1.0 / 8).astype(float),
                f"a_four_float={1.0}": (c == 1.0).astype(float),
            }
        )

        return input_df, {}, expected_df

    def data_with_missing(self):
        b, eb = self.binary_with_missing

        c = self._make_randn(len(b))

        input_df = pd.DataFrame({"a_binary": b, "a_number": c.copy()})

        expected_df = pd.DataFrame.from_dict({"a_binary=yes": eb, "a_number": c.copy()})

        return input_df, {}, expected_df

    def data_drop_all_missing(self):
        b, eb = self.binary_with_missing

        all_missing = pd.Series([np.nan] * len(b), dtype=object)

        input_df = pd.DataFrame({"a_binary": b, "bogus": all_missing})

        expected_df = pd.DataFrame({"a_binary=yes": eb})

        return input_df, {}, expected_df

    def data_retain_all_missing(self):
        input_df, _, expected_df = self.data_drop_all_missing()
        kwargs = {"allow_drop": False}
        expected_df.loc[:, "bogus"] = pd.Series([np.nan] * expected_df.shape[0], index=expected_df.index, dtype=object)

        return input_df, kwargs, expected_df

    def data_retain_only_one_level(self):
        b = np.r_[np.repeat(["yes"], 10)]

        input_df = pd.DataFrame({"categorical_col_with_only_one_level": b})
        expected_df = input_df.copy(deep=True)
        kwargs = {"allow_drop": False}

        return input_df, kwargs, expected_df


@pytest.mark.parametrize("inputs,kwargs,expected_df", EncodeCategoricalCases().get_cases())
@pytest.mark.filterwarnings(
    "ignore:In a future version, the Index constructor will not infer numeric dtypes when "
    "passed object-dtype sequences \\(matching Series behavior\\):FutureWarning"
)  # deprecated in pandas 1.4.0
def test_encode_categorical(inputs, kwargs, expected_df):
    actual_df = column.encode_categorical(inputs, **kwargs)
    tm.assert_frame_equal(actual_df.isnull(), expected_df.isnull())
    tm.assert_frame_equal(actual_df, expected_df, check_exact=True)


def test_series_numeric():
    input_series = pd.Series([0.5, 0.1, 10, 25, 3.8, 11, 2256, -1, -0.2, 3.14], name="a_series")

    with pytest.raises(TypeError, match="series must be of categorical dtype, but was float"):
        column.encode_categorical(input_series)


class CategoricalToNumeric(CategoricalCases):
    def data_categorical_series_to_numeric(self):
        input_series = pd.Series(
            ["a", "a", "b", "b", "b", "c"], name="Thr33", index=["Alpha", "Beta", "Gamma", "Delta", "Eta", "Mu"]
        )
        expected = pd.Series(
            [0, 0, 1, 1, 1, 2], name="Thr33", index=["Alpha", "Beta", "Gamma", "Delta", "Eta", "Mu"], dtype=np.int64
        )
        return input_series, expected

    def data_bool_series_to_numeric(self):
        input_series = pd.Series(
            [True, True, False, False, True, False, True],
            name="human",
            index=["Alpha", "Beta", "Gamma", "Delta", "Eta", "Mu", "Zeta"],
        )
        expected = pd.Series(
            [1, 1, 0, 0, 1, 0, 1],
            name="human",
            index=["Alpha", "Beta", "Gamma", "Delta", "Eta", "Mu", "Zeta"],
            dtype=np.int64,
        )
        return input_series, expected

    def data_frame_to_numeric(self):
        input_df = self.mixed_data_frame

        a_num = np.r_[np.repeat([0], 10), np.repeat([2], 5), np.repeat([3], 13), np.repeat([1], 3)].astype(np.int64)
        b_num = np.r_[np.repeat([1], 8), np.repeat([0], 23)].astype(np.int64)

        expected = pd.DataFrame.from_dict({"a_category": a_num, "a_binary": b_num})
        expected.loc[:, "a_number"] = input_df.loc[:, "a_number"].values.copy()

        return input_df, expected


@pytest.mark.parametrize("input_df,expected", CategoricalToNumeric().get_cases())
def test_categorical_to_numeric(input_df, expected):
    actual = column.categorical_to_numeric(input_df)

    if isinstance(expected, pd.Series):
        tm.assert_series_equal(actual, expected, check_exact=True)
    else:
        tm.assert_frame_equal(actual, expected, check_exact=True)



import numpy as np


def digitize_times(values: np.ndarray, time_step: float = 1.0) -> np.ndarray:
    """Generate unique time bin values that cover the input times
    and are rounded to the next time_step multiple

    :param values: an array times
    :param time_step: a step value for which the final bins will be based
    :return: an array of unique time bins that cover the input values
    """
    full_range = np.arange(
        _floor_base(np.min(values) - time_step, time_step), np.max(values) + time_step, step=time_step
    )
    times = full_range[np.digitize(values, full_range, right=True)]
    times = np.unique(times)
    return times


def stack_timepoints(X: np.ndarray, y: np.ndarray, times: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Generate a survival stacked dataset and the accompanying binary outcome
    for a survival dataset for all given timepoints.

    :param X: training input samples
    :param y: survival observations in the format of a 2d array, where
    the first column is the time and second column is the event
    :param times: array of time points on which to create risk sets
    :return: a tuple containing the survival stacked dataset and a binary
    outcome
    """
    num_times = times.shape[0]
    stacked_events = list(
        zip(*[_stack_timepoint(X, y, times, t) for t in np.arange(num_times)]))
    X_stacked = np.vstack(stacked_events[0])
    y_stacked = np.concatenate(stacked_events[1])
    return X_stacked, y_stacked


def _stack_timepoint(X: np.ndarray, y: np.ndarray, times: np.ndarray, i: int) -> tuple[np.ndarray, np.ndarray]:
    """Generate the predictor matrix and response vector for a survival dataset
    at a specific time-point `times[i]`.

    :param X: training input samples
    :param y: structured array with two fields. The binary event indicator
        as first field, and time of event or time of censoring as second field.
    :param times: array of time points on which to create risk sets
    :param i: index of array `times` at which to construct the dataset
    :return: a tuple containing the predictor matrix and response vector
    """
    event_field, time_field = y.dtype.names
    y_bins = np.digitize(y[time_field], times, right=True)
    X_i = X[y_bins >= i, :]
    y_i = y[y_bins >= i]
    X_risk = np.zeros((X_i.shape[0], times.shape[0]))
    X_risk[:, i] = 1
    y_outcome = ((np.digitize(y_i[time_field], times, right=True) == i) & (
        y_i[event_field])).astype(int)
    X_new = np.hstack((X_i, X_risk))
    return X_new, y_outcome


def stack_eval(X: np.ndarray, times: np.ndarray) -> np.ndarray:
    """Generate a predictor matrix for outcome prediction for given times. This
    is to be used for evaluation of a model, not for training.

    :param X: Survival input samples
    :param times: array of time points on which to create risk sets
    :return: a generalized predictor matrix for input X
    """
    X_cov = np.repeat(X, times.shape[0], axis=0)
    X_risk = np.tile(np.eye(times.shape[0]), (X.shape[0], 1))
    X_new = np.hstack((X_cov, X_risk))
    return X_new


def survival_function(estimates: np.ndarray, times: np.ndarray) -> np.ndarray:
    """Calculate the survival function from the stacked survival estimates.

    :param estimates: estimates as returned from a model trained on
    an evaluation set, shape (n_samples * n_timepoints,)
    :param times: array of time points on which the hazard was estimated,
    shape (n_timepoints,)
    :return: the survival function for each sample at the fitted time-points,
    shape (n_samples, n_timepoints)
    """
    # Reshape estimates to (n_samples, n_timepoints)
    est_matrix = estimates.reshape(-1, times.shape[0])

    # For each sample, compute survival function across timepoints
    surv = np.zeros_like(est_matrix)
    for i in range(est_matrix.shape[0]):
        # Calculate cumulative product for monotonic decrease
        surv[i, :] = np.cumprod(1 - est_matrix[i, :])

    return surv


def cumulative_hazard_function(estimates: np.ndarray, times: np.ndarray) -> np.ndarray:
    """Calculate the cumulative hazard function from the stacked survival
    estimates.

    :param estimates: estimates as returned from a model trained on
    an evaluation set
    :param times: array of time points on which the hazard was estimated
    :return: the cumulative hazard function for the fitted time-points
    """
    surv_curve = survival_function(estimates, times)
    surv_curve[surv_curve == 0] = np.finfo(float).eps
    return -np.log(surv_curve)


def risk_score(estimates: np.ndarray, times: np.ndarray) -> np.ndarray:
    """Calculate risk score from stacked survival estimates.

    :param estimates: estimates as returned from a model trained on
    an evaluation set
    :param times: array of time points on which the hazard was estimated
    :return: the risk score
    """
    chf = cumulative_hazard_function(estimates, times)
    return chf.sum(axis=1)


def _floor_base(x: float, base: float) -> float:
    """Return the floor of the input to the nearest multiple of given base

    :param x: input data
    :param base: the base of the floor calculations
    :return: the floor of the input data
    """
    return base * np.floor(x / base)
